#include <stdlib.h>
#include "utils.h"
#include "graph.h"
#include "linkedlist.h"

/* http://igm.univ-mlv.fr/~lecroq/string/bom.html#SECTION00245 */

#define FALSE      0
#define TRUE       1

int getTransition(char *x, int p, List L[], char c) {
  List cell;

  if (p > 0 && x[p - 1] == c)
    return(p - 1);
  else {
    cell = L[p];
    while (cell != NULL)
      if (x[cell->element] == c)
        return(cell->element);
      else
        cell = cell->next;
    return(UNDEFINED);
  }
}


void setTransition(int p, int q, List L[]) {
  List cell;

  cell = (List)malloc(sizeof(struct _cell));
  if (cell == NULL)
    error("BOM/setTransition");
  cell->element = q;
  cell->next = L[p];
  L[p] = cell;
}


void oracle(char *x, int m, char T[], List L[]) {
  int i, p, q;
  int S[XSIZE + 1];
  char c;

  S[m] = m + 1;
  for (i = m; i > 0; --i) {
    c = x[i - 1];
    p = S[i];
    while (p <= m &&
           (q = getTransition(x, p, L, c)) ==
           UNDEFINED) {
      setTransition(p, i - 1, L);
      p = S[p];
    }
    S[i - 1] = (p == m + 1 ? m : q);
  }
  p = 0;
  while (p <= m) {
    T[p] = TRUE;
    p = S[p];
  }
}

void BOM(char *x, int m, char *y, int n) {
  char T[XSIZE + 1];
  List L[XSIZE + 1];
  int i, j, p, period, q, shift;

  /* Preprocessing */
  memset(L, NULL, (m + 1)*sizeof(List));
  memset(T, FALSE, (m + 1)*sizeof(char));
  oracle(x, m, T, L);

  /* Searching */
  j = 0;
  while (j <= n - m) {
    i = m - 1;
    p = m;
    shift = m;
    /* while (i + j >= 0 && */
    /*        (q = getTransition(x, p, L, y[i + j])) != */
    /*        UNDEFINED) { */
    while (i + j >= 0 &&
           (q = getTransition(x, p, L, trace_get(y, i + j))) !=
           UNDEFINED) {
      p = q;
      if (T[p] == TRUE) {
        period = shift;
        shift = i;
      }
      --i;
    }
    if (i < 0) {
      OUTPUT(j);
      shift = period;
    }
    j += shift;
  }
}

List backward_oracle_matching(char* pattern, int pattern_size, 
                              char* text, int text_size) {
  return string_matcher_array(BOM, pattern, pattern_size, text, text_size);
}

int main(int argc, char* argv[]) {
  return string_matcher_main(BOM, argc, argv);
}
