#include "utils.h"
#include "graph.h"

/* http://igm.univ-mlv.fr/~lecroq/string/fdm.html#SECTION00220 */

void buildSuffixAutomaton(char *x, int m, Graph aut) {
  int i, art, init, last, p, q, r;
  char c;
  
  init = getInitial(aut);
  art = newVertex(aut);
  setSuffixLink(aut, init, art);
  last = init;
  for (i = 0; i < m; ++i) {
    c = x[i];
    p = last;
    q = newVertex(aut);
    setLength(aut, q, getLength(aut, p) + 1);
    setPosition(aut, q, getPosition(aut, p) + 1);
    while (p != init &&
           getTarget(aut, p, c) == UNDEFINED) {
      setTarget(aut, p, c, q);
      setShift(aut, p, c, getPosition(aut, q) -
               getPosition(aut, p) - 1);
      p = getSuffixLink(aut, p);
    }
    if (getTarget(aut, p, c) == UNDEFINED) {
      setTarget(aut, init, c, q);
      setShift(aut, init, c,
               getPosition(aut, q) -
               getPosition(aut, init) - 1);
      setSuffixLink(aut, q, init);
    }
    else
      if (getLength(aut, p) + 1 ==
          getLength(aut, getTarget(aut, p, c)))
        setSuffixLink(aut, q, getTarget(aut, p, c));
      else {
        r = newVertex(aut);
        copyVertex(aut, r, getTarget(aut, p, c));
        setLength(aut, r, getLength(aut, p) + 1);
        setSuffixLink(aut, getTarget(aut, p, c), r);
        setSuffixLink(aut, q, r);
        while (p != art &&
               getLength(aut, getTarget(aut, p, c)) >=
               getLength(aut, r)) {
          setShift(aut, p, c,
                   getPosition(aut,
                               getTarget(aut, p, c)) -
                   getPosition(aut, p) - 1);
          setTarget(aut, p, c, r);
          p = getSuffixLink(aut, p);
        }
      }
    last = q;
  }
  setTerminal(aut, last);
  while (last != init) {
    last = getSuffixLink(aut, last);
    setTerminal(aut, last);
  }
}


int FDM(char *x, int m, char *y, int n) {
  int j, init, ell, state;
  Graph aut;
 
  /* Preprocessing */
  aut = newSuffixAutomaton(2*(m + 2), 2*(m + 2)*ASIZE);
  buildSuffixAutomaton(x, m, aut);
  init = getInitial(aut);
 
  /* Searching */
  ell = 0;
  state = init;
  for (j = 0; j < n; ++j) {
    /* if (getTarget(aut, state, y[j]) != UNDEFINED) { */
    if (getTarget(aut, state, trace_get(y, j)) != UNDEFINED) {
      ++ell;
      state = getTarget(aut, state, y[j]);
    }
    else {
      while (state != init &&
             getTarget(aut, state, y[j]) == UNDEFINED)
        state = getSuffixLink(aut, state);
      if (getTarget(aut, state, y[j]) != UNDEFINED) {
        ell = getLength(aut, state) + 1;
        state = getTarget(aut, state, y[j]);
      }
      else {
        ell = 0;
        state = init;
      }
    }
    if (ell == m)
      OUTPUT(j - m + 1);
  }
}

List forward_dawg_matching(char* pattern, int pattern_size, 
                           char* text, int text_size) {
  return string_matcher_array(FDM, pattern, pattern_size, text, text_size);
}

int main(int argc, char* argv[]) {
  return string_matcher_main(FDM, argc, argv);
}
