#include "utils.h"

/* http://igm.univ-mlv.fr/~lecroq/string/node29.html#SECTION00290 */

typedef struct patternScanOrder {
  int loc;
  char c;
} pattern;

/* Construct an ordered pattern from a string. */
void orderPattern(char *x, int m, int (*pcmp)(),
                  pattern *pat) {
  int i;

  for (i = 0; i <= m; ++i) {
    pat[i].loc = i;
    pat[i].c = x[i];
  }
  qsort(pat, m, sizeof(pattern), pcmp);
}

/* Find the next leftward matching shift for
   the first ploc pattern elements after a
   current shift or lshift. */
int matchShift(char *x, int m, int ploc,
               int lshift, pattern *pat) {
  int i, j;

  for (; lshift < m; ++lshift) {
    i = ploc;
    while (--i >= 0) {
      if ((j = (pat[i].loc - lshift)) < 0)
        continue;
      if (pat[i].c != x[j])
        break;
    }
    if (i < 0)
      break;
  }
  return(lshift);
}

/* Constructs the good-suffix shift table
   from an ordered string. */
void preAdaptedGs(char *x, int m, int adaptedGs[],
                  pattern *pat) {
  int lshift, i, ploc;

  adaptedGs[0] = lshift = 1;
  for (ploc = 1; ploc <= m; ++ploc) {
    lshift = matchShift(x, m, ploc, lshift, pat);
    adaptedGs[ploc] = lshift;
  }
  for (ploc = 0; ploc <= m; ++ploc) {
    lshift = adaptedGs[ploc];
    while (lshift < m) {
      i = pat[ploc].loc - lshift;
      if (i < 0 || pat[ploc].c != x[i])
        break;
      ++lshift;
      lshift = matchShift(x, m, ploc, lshift, pat);
    }
    adaptedGs[ploc] = lshift;
  }
}

int minShift[XSIZE];

/* Computation of the MinShift table values. */
void computeMinShift(char *x, int m) {
  int i, j;

  for (i = 0; i < m; ++i) {
    for (j = i - 1; j >= 0; --j)
      if (x[i] == x[j])
        break;
    minShift[i] = i - j;
  }
}


/* Maximal Shift pattern comparison function. */
int maxShiftPcmp(pattern *pat1, pattern *pat2) {
  int dsh;

  dsh = minShift[pat2->loc] - minShift[pat1->loc];
  return(dsh ? dsh : (pat2->loc - pat1->loc));
}

void preQsBc(char *x, int m, int qsBc[]) {
  int i;

  for (i = 0; i < ASIZE; ++i)
    qsBc[i] = m + 1;
  for (i = 0; i < m; ++i)
    qsBc[x[i]] = m - i;
}

/* Maximal Shift string matching algorithm. */
void MS(char *x, int m, char *y, int n) {
  int i, j, qsBc[ASIZE], adaptedGs[XSIZE];
  pattern pat[XSIZE];

  /* Preprocessing */
  computeMinShift(x ,m);
  orderPattern(x, m, maxShiftPcmp, pat);
  preQsBc(x, m, qsBc);
  preAdaptedGs(x, m, adaptedGs, pat);

  /* Searching */
  j = 0;
  set_txt_length(n);
  while (j <= n - m) {
    start_pruning_duplicates();
    i = 0;
    while (i < m && pat[i].c == trace_get(y, j + pat[i].loc))
      ++i;
    if (i >= m)
      OUTPUT(j);
    j += MAX(adaptedGs[i], qsBc[trace_get(y, j + m)]);
    stop_pruning_duplicates();

    /* i = 0; */
    /* while (i < m && pat[i].c == y[j + pat[i].loc]) */
    /*   ++i; */
    /* if (i >= m) */
    /*   OUTPUT(j); */
    /* j += MAX(adaptedGs[i], qsBc[y[j + m]]); */
  }
}

List maximal_shift(char* pattern, int pattern_size, 
                   char* text, int text_size) {
  return string_matcher_array(MS, pattern, pattern_size, text, text_size);
}

int main(int argc, char* argv[]) {
  return string_matcher_main(MS, argc, argv);
}
