#include <stdio.h>
#include <string.h>
#include "utils.h"

/*
   Helmuth Partsch and Frank A. Stomp. A fast pattern matching
   algorithm derived by transformational and assertional
   reasoning. Formal Aspects of Com- puting, 2(2):109–122, 1990
 */

int delta(char *x, int m, char c, int l) {
  int i;

  /* ^^ := sequential conjunuction (mapand) */

  /* funct delta = (char c, nat l) nat:
       |p| - max { nat i: (|p| - l < i < |p| ^^
                            (p[i - |p| + l + 1 : i] = p[l + 1 : |p|]
                              ^ p[i - |p| + l] = c))
                          OR (0 <= i <= |p| - 1 ^^ p[1 : i] =
                               p[|p| - i + 1 : |p|]) }
  */

/*
3 2 4 6 10 9 8 7

01234567890
abbabacabaa
abaa
  !=
 abaa
    !
   abaa
      !
     abaa
        !
       abaa
       ====
 */

  printf("\ndelta(%c, %d)\n", c, l);

  for(i = m - 1; i >= 0; i--) {
    printf("i = %d\n", i);
    if (i < m && i > m - l) {
      printf("first: %s = %s (%d) => %d\n",
             x + (i - m + l), 
             x + l, 
             m - l,
             (memcmp(x + (i - m + l),
                     x + l,
                     m - l)));
      if (memcmp(x + (i - m + l),
                 x + l,
                 m-l) == 0
          && x[i - m + l - 1] == c) {
        printf("!");
        return m - i;
      }
    }
    if (i < m - 1 && i >= 0) {
      if (memcmp(x,
                 x + (m - i + 1),
                 i) == 0) {
        printf("@");
        return m - i;
      }
    }
  }

  printf("#");
  return m;
}

int d(char *x, int m, char c, int l) {
  /* d[c, l] = |p| - l + \delta(c, l) */
  return m - 1 - l + delta(x, m, c, l);
}

void PS(char *x, int m, char *y, int n) {
  int k, l;

  /*
     funct find = (string p, string s) res:
       find' (|p|, |p|) where
       funct find' = (nat k, nat l) res:
         if k > |s| then false
         elsf s[k] != p[|p|] then
           find'(k + d[s[k], |p|], |p|)
         else
           try-match'(k-1, l-1)
         fi;
       funct try-match' = (nat k, nat l) res:
         if l = 0 then
           k + 1
         elsf s[k] == p[l] then
           try-match'(k-1, l-1)
         else
           find'(k + d[s[k], l], |p|)
         fi
  */

  k = l = m - 1;
  while (k < n) {
    if (l < 0) {
      OUTPUT(k + 1);
    }
    /* if (x[l] != y[k]) { */
    if (x[l] != trace_get(y, k)) {
      printf(" pat[%d]=%c, txt[%d]=%c", l, x[l], k, y[k]);
      printf(" d[%c, %d] = %d\n", y[k], l, d(x, m, y[k], l));
      k = k + d(x, m, y[k], l);
      l = m - 1;
    } else {
      printf(" pat[%d]=%c, txt[%d]=%c", l, x[l], k, y[k]);
      printf("\n");
      k--;
      l--;
    }
  }
}

/*
3 2 4 6 10 9 8 7

01234567890
abbabacabaa
abaa
  !=
 abaa
    !
   abaa
      !
     abaa
        !
       abaa
       ====
 */

int main(int argc, char* argv[]) {
  return string_matcher_main(PS, argc, argv);
}
