#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "linkedlist.h"
#include "utils.h"

List trace_first = NULL;
List trace_current = NULL;

int txt_length = -1;

int prune_duplicates = 0;
List prune_from_trace = NULL;

void set_txt_length(int n) {
  txt_length = n;
}

void start_pruning_duplicates() {
  prune_duplicates = 1;
  prune_from_trace = NULL;
}
void stop_pruning_duplicates() {
  prune_duplicates = 0;
}

void trace_new() {
  List tmp, tmp2;
  tmp = trace_first;
  while (NULL != tmp) {
    tmp2 = tmp;
    tmp = tmp->next;
    free(tmp2);
  }

  trace_first = trace_current = NULL;
}

void trace_add(int j) {
  List trace_entry;

  trace_entry = (List)malloc(sizeof(struct _cell));
  trace_entry->element = j;
  trace_entry->next = NULL;

  if (NULL == trace_first) {
    trace_first = trace_entry;
    trace_current = trace_first;
  } else {
    trace_current->next = trace_entry;
    trace_current = trace_entry;
  }  
}

void trace_print() {
  List e;

  e = trace_first;
  while(NULL != e) {
    printf("%d ", e->element);
    e = e->next;
  }
  printf("\n");
}

void trace(int j) {
  /* Dont trace txt indices outside range */
  if (j < 0 || (txt_length > 0 && j >= txt_length)) {
    return;
  }

  /* if enabled, don't record duplicate indices */
  if (prune_duplicates) {
    List tmp;
    tmp = prune_from_trace;
    while (NULL != tmp) {
      if (tmp->element == j) {
        return;
      }
      tmp = tmp->next;
    }
  }

  trace_add(j);

  /* don't record duplicates indices starting from here */
  if (prune_duplicates && NULL == prune_from_trace) {
    prune_from_trace = trace_current;
  }
}

char trace_get(char *txt, int j) {
  trace(j);
  return txt[j];
}

int trace_memcmp(char *pat, char *txt, int length, int txt_offset) {
  int i;
  for(i = 0; i < length; i++) {
    trace(txt_offset + i);
    if (pat[i] != txt[i])
      return 1;
  }
  return 0;
}

int string_matcher_main(void (*matcher)(char*, int, char*, int), 
                        int argc, char* argv[]) {
  if (argc < 2) {
    printf("%s pattern text\n", argv[0]);
    return 2;
  }

  if (strlen(argv[1]) >= XSIZE) {
    error("Pattern too large.");
  }

  matcher(argv[1], strlen(argv[1]), argv[2], strlen(argv[2]));
  trace_print();
  return 0;
}

void error(char* str) {
  printf("Error: %s\n", str);
  exit(1);
}

int MAX(int a, int b) {
  return (a > b ? a : b);
}

int MIN(int a, int b) {
  return (a < b ? a : b);
}

List string_matcher_array(void (*matcher)(char*, int, char*, int), 

                          char* pattern, int pattern_size, 
                          char* text, int text_size) {
  trace_new();
  matcher(pattern, pattern_size, text, text_size);
  return trace_first;
}
