
(define naive (make-matcher (match-basic order-left-to-right prune-all)))

(define mp (make-matcher (match-basic order-left-to-right prune-neg)))

(define kmp (make-matcher (match-basic order-left-to-right 
                                       (prune-neg-older-than 1))))

(define horspool 
  (make-matcher
   (match-backtracking
    (match-basic order-last-left-to-right prune-all)
    (match-table-shifts order-last-only (prune-older-than 1)))))

(define horspool-right-to-left
  (make-matcher
   (match-backtracking
    (match-basic order-right-to-left prune-all)
    (match-table-shifts order-last-only (prune-older-than 1)))))

(define boyer-moore
  (make-matcher
   (match-skew
    (match-basic-shifts order-right-to-left (prune-older-than 1))
    (match-table-shifts order-last-only (prune-older-than 1)))))

(define automaton
  (make-matcher
   (match-table order-left-to-right prune-none)))

(define optimal-bm
  (make-matcher
   (match-table order-right-to-left prune-none)))

(define partsch-stomp
  (make-matcher
   (match-table-shifts order-right-to-left (prune-older-than 1))))

(define original-bm
  (make-matcher
   (match-skew
    (match-basic-shifts order-right-to-left (lambda (cache)
                                              ((prune-older-than 1)
                                               (prune-neg cache))))
    (match-table-shifts order-last-only (prune-older-than 1)))))

(define raita
  (make-matcher
   (match-backtracking
    (match-basic order-last-first-middle-rest prune-all)
    (match-table-shifts order-last-only (prune-older-than 1)))))

(define quick-search 
  (make-matcher
   (match-alternate
    (match-basic order-left-to-right prune-all)
    (match-table-shifts order-last-only (prune-older-than 1)))))

(define smith
  (make-matcher
   (match-backtracking
    (match-basic order-left-to-right prune-all)
    (match-parallel
     (match-table-shifts order-last-only (prune-older-than 1))
     (match-alternate
      (match-fail)
      (match-table-shifts order-last-only (prune-older-than 1)))))))

(define not-so-naive
  (make-matcher
   (match-sequential
    (match-basic-shifts order-second-only (prune-older-than 1))
    (match-basic order-third-to-right-first prune-all))))

(define zhu-takaoka
  (make-matcher
   (match-backtracking
    (match-basic-shifts order-right-to-left (prune-older-than 1))
    (match-table-full-shifts order-last-two (prune-older-than 1)))))

(define matchers0
  `(
    ("naive" ,naive)
    ("mp" ,mp)
    ("kmp" ,kmp)
    ;; ("neg2" ,posneg2)
    ;; ("negonly" ,negonly)
    ;; ("pos1" ,pos1)
    ;; ("posneg1" ,posneg1)
    ;; ("posneg2" ,posneg2)
    ))

(define pat0 "abaa")
(define txt0 "abbabacabaa")

(define pat1 "cababa")
(define txt1 "xxxxaababacababa")

(define (run-tests testcases)
  (if (null? testcases)
      '()
      (let ((test (car testcases)))
        (if (equal? ((list-ref test 1) 
                     (list-ref test 2)
                     (list-ref test 3))
                    (list-ref test 4))
            (run-tests (cdr testcases))
            (cons (list-ref test 0)
                  (run-tests (cdr testcases)))))))

(define testcases
  `(("naive0" ,naive ,pat0 ,txt0 (0 1 2 1 2 3 4 5 6 4 5 6 6 7 8 9 10))
    ("mp0" ,mp ,pat0 ,txt0 (0 1 2 2 3 4 5 6 6 6 7 8 9 10))
    ("kmp0" ,kmp ,pat0 ,txt0 (0 1 2 3 4 5 6 6 6 7 8 9 10))
    ("automaton0" ,automaton ,pat0 ,txt0 (0 1 2 3 4 5 6 7 8 9 10))
    ("optimal-bm0" ,optimal-bm ,pat0 ,txt0 (3 2 4 6 10 9 8 7))
    ("optimal-bm1" ,optimal-bm "aa" "baa" (1 0 2))
    ("partsch-stomp0" ,partsch-stomp ,pat0 ,txt0 (3 2 4 6 10 9 8 7))
    ("partsch-stomp1" ,partsch-stomp "aa" "baa" (1 0 2 1))
    ("boyer-moore0" ,boyer-moore ,pat0 ,txt0 (3 2 4 6 10 9 8 7))
    ("boyer-moore1" ,boyer-moore ,pat1 ,txt1 (5 4 9 8 7 6 5 4 15 14 13 12 11 10))
    ("original-bm0" ,original-bm ,pat0 ,txt0 (3 2 4 6 10 9 8 7))
    ("original-bm1" ,original-bm ,pat1 ,txt1 
     (5 4 7 6 5 4 9 8 7 6 5 4 15 14 13 12 11 10))
    ("raita0" ,raita ,pat0 ,txt0 (3 0 2 4 6 10 7 9 8))
    ("quick-search0" ,quick-search ,pat0 ,txt0 (0 1 2 4 3 4 5 6 7 4 8 7 8 9 10))
    ("smith0" ,smith ,pat0 ,txt0 (0 1 2 3 4 3 4 5 6 7 7 8 9 10))
    ("not-so-naive0" ,not-so-naive ,pat0 ,txt0 (1 2 3 4 5 6 6 7 8 9 10 7))
    ))

(define (test)
  (run-tests testcases))

(define (ager-al-2002-ASIA-PEPM02-fig-3 pat txt)
  "Mads Sig Ager, Olivier Danvy, and Henning Korsholm Rohde. On
obtaining Knuth, Morris, and Pratt’s string matcher by partial evalu-
ation. In Wei-Ngan Chin, editor, Proceedings of the 2002 ACM SIG- PLAN
Asian Symposium on Partial Evaluation and Semantics-Based Program
Manipulation (ASIA-PEPM 2002), pages 32–46, Aizu, Japan, September
2002. ACM Press."
  (define trace '())
  (define (traced-equal? char k) 
    (set! trace (cons k trace)) 
    (equal? char (string-ref txt k)))
  (let ([lpat (string-length pat)] [ltxt (string-length txt)])
    (letrec ([match
              (lambda (j k)
                (if (= j lpat)
                    (- k j)
                    (if (= k ltxt)
                        -1
                        (compare j k))))]
             [compare
              (lambda (j k)
                ;; (if (eq? (string-ref pat j)
                ;;          (string-ref txt k))
                (if (traced-equal? (string-ref pat j) k)
                    (match (+ j 1) (+ k 1))
                    (if (= 0 j)
                        (match 0 (+ k 1))
                        (rematch j k 0 1))))]
             [rematch
              (lambda (j k jp kp)
                (if (= kp j)
                    (if (eq? (string-ref pat jp)
                             (string-ref pat kp))
                        (if (= jp 0)
                            (match 0 (+ k 1))
                            (rematch j k 0 (+ (- kp jp) 1)))
                        (compare jp k))
                    (if (eq? (string-ref pat jp)
                             (string-ref pat kp))
                        (rematch j k (+ jp 1) (+ kp 1))
                        (rematch j k 0 (+ (- kp jp) 1)))))])
      (match 0 0)
      (reverse trace))))

(define (ager-al-2002-ASIA-PEPM02-fig-6 pat txt)
  "Mads Sig Ager, Olivier Danvy, and Henning Korsholm Rohde. On
obtaining Knuth, Morris, and Pratt’s string matcher by partial evalu-
ation. In Wei-Ngan Chin, editor, Proceedings of the 2002 ACM SIG- PLAN
Asian Symposium on Partial Evaluation and Semantics-Based Program
Manipulation (ASIA-PEPM 2002), pages 32–46, Aizu, Japan, September
2002. ACM Press."
  (define trace '())
  (define (traced-equal? char k) 
    (set! trace (cons k trace)) 
    (equal? char (string-ref txt k)))
  (let ([lpat (string-length pat)] [ltxt (string-length txt)])
    (letrec ([match
              (lambda (j k)
                (if (= j lpat)
                    (- k j)
                    (if (= k ltxt)
                        -1
                        (compare j k))))]
             [compare
              (lambda (j k)
                ;; (if (eq? (string-ref pat j)
                ;;          (string-ref txt k))
                (if (traced-equal? (string-ref pat j) k)
                    (match (+ j 1) (+ k 1))
                    (if (= 0 j)
                        (match 0 (+ k 1))
                        (let ([next (rematch j 0 1)])
                          (if (= next -1)
                              (match 0 (+ k 1))
                              (compare next k))))))]
             [rematch
              (lambda (j jp kp)
                (if (= kp j)
                    (if (eq? (string-ref pat jp)
                             (string-ref pat kp))
                        (if (= jp 0)
                            -1
                            (rematch j 0 (+ (- kp jp) 1)))
                        jp)
                    (if (eq? (string-ref pat jp)
                             (string-ref pat kp))
                        (rematch j (+ jp 1) (+ kp 1))
                        (rematch j 0 (+ (- kp jp) 1)))))])
      (match 0 0)
      (reverse trace))))


(define (ager-al-TOPLAS06-fig-1 pat txt)
  "Mads Sig Ager, Olivier Danvy, and Henning Korsholm Rohde. Fast
partial evaluation of pattern matching in strings. ACM Transactions on
Programming Languages and Systems, 28(4):696–714, July 2006. A
preliminary version was presented at the 2003 ACM SIGPLAN Work- shop
on Partial Evaluation and Semantics-Based Program Manipula- tion (PEPM
2003)."
  (define trace '())
  (define (traced-equal? char k) 
    (set! trace (cons k trace)) 
    (equal? char (string-ref txt k)))
  (define (main pattern text)
    (match pattern text 0 0))
  (define (match pattern text j k)
    (if (= (string-length pattern) j)
        (- k j)
        (if (= (string-length text) k)
            -1
            (compare pattern text j k))))
  (define (compare pattern text j k)
    ;; (if (equal? (string-ref pattern j)
    ;;             (string-ref text k))
    (if (traced-equal? (string-ref pattern j) k)
        (match pattern text (+ j 1) (+ k 1))
        (let ([s (rematch pattern j)])
          (if (= s -1)
              (match pattern text 0 (+ k 1))
              (compare pattern text s k)))))
  (define (rematch pattern i)
    (if (= i 0)
        -1
        (letrec ([try (lambda (jp kp)
                        (if (= kp i)
                            jp
                            (if (equal? (string-ref pattern jp)
                                        (string-ref pattern kp))
                                (try (+ jp 1) (+ kp 1))
                                (try 0 (+ (- kp jp) 1)))))])
          (try 0 1))))
  (main pat txt)
  (reverse trace))

(define (ager-al-TOPLAS06-fig-3 pat txt)
  "Mads Sig Ager, Olivier Danvy, and Henning Korsholm Rohde. Fast
partial evaluation of pattern matching in strings. ACM Transactions on
Programming Languages and Systems, 28(4):696–714, July 2006. A
preliminary version was presented at the 2003 ACM SIGPLAN Work- shop
on Partial Evaluation and Semantics-Based Program Manipula- tion (PEPM
2003)."
  (define trace '())
  (define (traced-equal? char k) 
    (set! trace (cons k trace)) 
    (equal? char (string-ref txt k)))
  (define (main pattern text)
    (match pattern text 0 0))
  (define (match pattern text j k)
    (if (= (string-length pattern) j)
        (- k j)
        (if (= (string-length text) k)
            -1
            (compare pattern text j k))))
  (define (compare pattern text j k)
    ;; (if (equal? (string-ref pattern j)
    ;;             (string-ref text k))
    (if (traced-equal? (string-ref pattern j) k)
        (match pattern text (+ j 1) (+ k 1))
        (let ([s (rematch pattern j)])
          (if (= s -1)
              (match pattern text 0 (+ k 1))
              (compare pattern text s k)))))
  (define (rematch pattern i)
    (if (= i 0)
        -1
        (letrec ([try-subproblem
                  (lambda (j)
                    (if (= j -1)
                        0
                        (if (equal? (string-ref pattern j)
                                    (string-ref pattern (- i 1)))
                            (+ j 1)
                            (try-subproblem (rematch pattern j)))))])
          (try-subproblem (rematch pattern (- i 1))))))
  (main pat txt)
  (reverse trace))


(define (ager-al-TOPLAS06-fig-4 pat txt)
  "Mads Sig Ager, Olivier Danvy, and Henning Korsholm Rohde. Fast
partial evaluation of pattern matching in strings. ACM Transactions on
Programming Languages and Systems, 28(4):696–714, July 2006. A
preliminary version was presented at the 2003 ACM SIGPLAN Work- shop
on Partial Evaluation and Semantics-Based Program Manipula- tion (PEPM
2003)."
  (define trace '())
  (define (traced-equal? char k) 
    (set! trace (cons k trace)) 
    (equal? char (string-ref txt k)))
  (define (main pattern text)
    (match pattern text 0 0))
  (define (match pattern text j k)
    (if (= (string-length pattern) j)
        (- k j)
        (if (= (string-length text) k)
            -1
            (compare pattern text j k))))
  (define (compare pattern text j k)
    ;; (if (equal? (string-ref pattern j)
    ;;             (string-ref text k))
    (if (traced-equal? (string-ref pattern j) k)
        (match pattern text (+ j 1) (+ k 1))
        (let ([s (rematch-neg pattern j)])
          (if (= s -1)
              (match pattern text 0 (+ k 1))
              (compare pattern text s k)))))
  (define (rematch-neg pattern i)
    (if (= i 0)
        -1
        (let ([j (rematch pattern i)])
          (if (equal? (string-ref pattern j)
                      (string-ref pattern i))
              (rematch-neg pattern j)
              j))))
  (define (rematch pattern i)
    (if (= i 0)
        -1
        (letrec ([try-subproblem
                  (lambda (j)
                    (if (= j -1)
                        0
                        (if (equal? (string-ref pattern j)
                                    (string-ref pattern (- i 1)))
                            (+ j 1)
                            (try-subproblem (rematch pattern j)))))])
          (try-subproblem (rematch pattern (- i 1))))))
  (main pat txt)
  (reverse trace))

(define (amtoft-al-Jones02-left-to-right pat txt)
  "Torben Amtoft, Charles Consel, Olivier Danvy, and Karoline
2Malmkjær. The abstraction and instantiation of string-matching pro-
grams. In Torben Æ. Mogensen, David A. Schmidt, and I. Hal Sud-
borough, editors, The Essence of Computation: Complexity, Analysis,
Transformation. Essays Dedicated to Neil D. Jones, number 2566 in
Lecture Notes in Computer Science, pages 332–357. Springer, 2002."
  (define trace '())
  (define (traced-equal? char k) 
    (set! trace (cons k trace)) 
    (equal? char (string-ref txt k)))
  (define pos? char?)
  (define (neg? e)
    (or (null? e) (pair? e)))
  (define (cache-init n)
    (if (= n 0)
        '()
         (cons '() (cache-init (- n 1)))))
  (define (cache-shift c)
    (append (cdr c) (list '())))
  (define (cache-ref-pos? c i)
    (pos? (list-ref c i)))
  (define (cache-ref-neg? c i)
    (neg? (list-ref c i)))
  (define cache-ref-pos list-ref)
  (define cache-ref-neg list-ref)
  (define (cache-extend-pos c i pos)
    (letrec ([walk
              (lambda (c i)
                (if (= i 0)
                    (cons pos (cdr c))
                    (cons (car c) (walk (cdr c) (- i 1)))))])
      (walk c i)))
  (define (cache-extend-neg c i neg)
    (letrec ([walk
              (lambda (c i)
                (if (= i 0)
                    (cons (cons neg (car c)) (cdr c))
                    (cons (car c) (walk (cdr c) (- i 1)))))])
      (walk c i)))
  (define (cache-prune c)
    c)
  (define (schedule-pc c)
    (letrec ([walk
              (lambda (i c)
                (cond
                 [(null? c)
                  '()]
                 [(pos? (car c))
                  (cons i (walk (+ i 1) (cdr c)))]
                 [else
                  (if (null? (car c))
                      (walk (+ i 1) (cdr c))
                      (cons i (walk (+ i 1) (cdr c))))]))])
      (walk 0 c)))
  (define (schedule-pt c)
    (letrec ([walk
              (lambda (i c)
                (cond
                 [(null? c)
                  '()]
                 [(pos? (car c))
                  (walk (+ i 1) (cdr c))]
                 [else
                  (cons i (walk (+ i 1) (cdr c)))]))])
      (walk 0 c)))
  (define (match p t)
    (let ([lp (string-length p)])       ;;; lp is the length of p
      (if (= lp 0)
          0
          (let ([lt (string-length t)]) ;;; lt is the length of t
            (if (< lt lp)
                -1
                (match-pt p lp (cache-init lp) t 0 lt))))))
  (define (match-pc p lp c t bt lt) 
    ;;; matches p[0..lp-1] and c[0..lp-1]
    (letrec ([loop-pc
              (lambda (z c is)
                (if (null? is)
                    (let ([bt (+ bt z)] [lt (- lt z)])
                      (if (< lt lp)
                          -1
                          (match-pt p lp c t bt lt)))
                    (let ([i (car is)])
                      (if (if (cache-ref-pos? c i)
                              (equal? (string-ref p i)
                                      (cache-ref-pos c i))
                              (not (member (string-ref p i)
                                           (cache-ref-neg c i))))
                          (loop-pc z c (cdr is))
                          (let ([c (cache-shift c)])
                            (loop-pc (+ z 1) c (schedule-pc c)))))))])
      (loop-pc 1 c (schedule-pc c))))
  (define (match-pt p lp c t bt lt) 
    ;;; matches p[0..lp-1] and t[bt..bt+lp-1]
    (letrec ([loop-pt               ;;; bt is the base index of t
              (lambda (c is)
                (if (null? is)
                    bt
                    (let* ([i (car is)] [x (string-ref p i)])
                      ;; (if (equal? (string-ref t (+ bt i)) x)
                      (if (traced-equal? x (+ bt i))
                          (loop-pt (cache-prune
                                    (cache-extend-pos c i x))
                                   (cdr is))
                          (match-pc p lp
                                    (cache-shift
                                     (cache-extend-neg c i x))
                                    t bt lt)))))])
      (loop-pt c (schedule-pt c))))
  (match pat txt)
  (reverse trace))

(define (amtoft-al-Jones02-right-to-left pat txt)
  "Torben Amtoft, Charles Consel, Olivier Danvy, and Karoline
Malmkjær. The abstraction and instantiation of string-matching pro-
grams. In Torben Æ. Mogensen, David A. Schmidt, and I. Hal Sud-
borough, editors, The Essence of Computation: Complexity, Analysis,
Transformation. Essays Dedicated to Neil D. Jones, number 2566 in
Lecture Notes in Computer Science, pages 332–357. Springer, 2002."
  (define trace '())
  (define (traced-equal? char k) 
    (set! trace (cons k trace)) 
    (equal? char (string-ref txt k)))
  (define pos? char?)
  (define (neg? e)
    (or (null? e) (pair? e)))
  (define (cache-init n)
    (if (= n 0)
        '()
         (cons '() (cache-init (- n 1)))))
  (define (cache-shift c)
    (append (cdr c) (list '())))
  (define (cache-ref-pos? c i)
    (pos? (list-ref c i)))
  (define (cache-ref-neg? c i)
    (neg? (list-ref c i)))
  (define cache-ref-pos list-ref)
  (define cache-ref-neg list-ref)
  (define (cache-extend-pos c i pos)
    (letrec ([walk
              (lambda (c i)
                (if (= i 0)
                    (cons pos (cdr c))
                    (cons (car c) (walk (cdr c) (- i 1)))))])
      (walk c i)))
  (define (cache-extend-neg c i neg)
    (letrec ([walk
              (lambda (c i)
                (if (= i 0)
                    (cons (cons neg (car c)) (cdr c))
                    (cons (car c) (walk (cdr c) (- i 1)))))])
      (walk c i)))
  (define (cache-prune c)
    c)
  (define (schedule-pc c)
    (letrec ([walk
              (lambda (i c)
                (cond
                 [(null? c)
                  '()]
                 [(pos? (car c))
                  (cons i (walk (+ i 1) (cdr c)))]
                 [else
                  (if (null? (car c))
                      (walk (+ i 1) (cdr c))
                      (cons i (walk (+ i 1) (cdr c))))]))])
      (reverse (walk 0 c))))
  (define (schedule-pt c)
    (letrec ([walk
              (lambda (i c)
                (cond
                 [(null? c)
                  '()]
                 [(pos? (car c))
                  (walk (+ i 1) (cdr c))]
                 [else
                  (cons i (walk (+ i 1) (cdr c)))]))])
      (reverse (walk 0 c))))
  (define (match p t)
    (let ([lp (string-length p)])       ;;; lp is the length of p
      (if (= lp 0)
          0
          (let ([lt (string-length t)]) ;;; lt is the length of t
            (if (< lt lp)
                -1
                (match-pt p lp (cache-init lp) t 0 lt))))))
  (define (match-pc p lp c t bt lt) 
    ;;; matches p[0..lp-1] and c[0..lp-1]
    (letrec ([loop-pc
              (lambda (z c is)
                (if (null? is)
                    (let ([bt (+ bt z)] [lt (- lt z)])
                      (if (< lt lp)
                          -1
                          (match-pt p lp c t bt lt)))
                    (let ([i (car is)])
                      (if (if (cache-ref-pos? c i)
                              (equal? (string-ref p i)
                                      (cache-ref-pos c i))
                              (not (member (string-ref p i)
                                           (cache-ref-neg c i))))
                          (loop-pc z c (cdr is))
                          (let ([c (cache-shift c)])
                            (loop-pc (+ z 1) c (schedule-pc c)))))))])
      (loop-pc 1 c (schedule-pc c))))
  (define (match-pt p lp c t bt lt) 
    ;;; matches p[0..lp-1] and t[bt..bt+lp-1]
    (letrec ([loop-pt               ;;; bt is the base index of t
              (lambda (c is)
                (if (null? is)
                    bt
                    (let* ([i (car is)] [x (string-ref p i)])
                      ;; (if (equal? (string-ref t (+ bt i)) x)
                      (if (traced-equal? x (+ bt i))
                          (loop-pt (cache-prune
                                    (cache-extend-pos c i x))
                                   (cdr is))
                          (match-pc p lp
                                    (cache-shift
                                     (cache-extend-neg c i x))
                                    t bt lt)))))])
      (loop-pt c (schedule-pt c))))
  (match pat txt)
  (reverse trace))

(define (amtoft-al-Jones02-fig-2-aaa pat txt)
  "Torben Amtoft, Charles Consel, Olivier Danvy, and Karoline
Malmkjær. The abstraction and instantiation of string-matching pro-
grams. In Torben Æ. Mogensen, David A. Schmidt, and I. Hal Sud-
borough, editors, The Essence of Computation: Complexity, Analysis,
Transformation. Essays Dedicated to Neil D. Jones, number 2566 in
Lecture Notes in Computer Science, pages 332–357. Springer, 2002."
  (define trace '())
  (define (traced-equal? char k) 
    (set! trace (cons k trace)) 
    (equal? char (string-ref txt k)))
  (define (match-aaa t)
    (define (loop-pc-1 bt lt z)
      (let ([bt (+ bt z)] [lt (- lt z)])
        (if (< lt 3)
            -1
            (loop-pt-1 bt lt))))
    (define (loop-pt-1 bt lt)
      ;; (if (equal? (string-ref t (+ bt 0)) #\a)
      (if (traced-equal? #\a (+ bt 0))
          ;; (if (equal? (string-ref t (+ bt 1)) #\a)
          (if (traced-equal? #\a (+ bt 1))
              ;; (if (equal? (string-ref t (+ bt 2)) #\a)
              (if (traced-equal? #\a (+ bt 2))
                  bt
                  (loop-pc-1 bt lt 3))
              (loop-pc-1 bt lt 2))
          (loop-pc-1 bt lt 1)))
    (let ([lt (string-length t)])
      (if (< lt 3)
          -1
          (loop-pt-1 0 lt))))
  (if (string=? pat "aaa")
      (begin (match-aaa txt)
             (reverse trace))
      '()))

(define (amtoft-al-Jones02-fig-3-abb pat txt)
  "Torben Amtoft, Charles Consel, Olivier Danvy, and Karoline
Malmkjær. The abstraction and instantiation of string-matching pro-
grams. In Torben Æ. Mogensen, David A. Schmidt, and I. Hal Sud-
borough, editors, The Essence of Computation: Complexity, Analysis,
Transformation. Essays Dedicated to Neil D. Jones, number 2566 in
Lecture Notes in Computer Science, pages 332–357. Springer, 2002."
  (define trace '())
  (define (traced-equal? char k) 
    (set! trace (cons k trace)) 
    (equal? char (string-ref txt k)))
  (define (match-abb t)
    (define (loop-pc-1 bt lt z)
      (let ([bt (+ bt z)] [lt (- lt z)])
        (if (< lt 3)
            -1
            (loop-pt-1 bt lt))))
    (define (loop-pt-1 bt lt)
      ;; (if (equal? (string-ref t (+ bt 2)) #\b)
      (if (traced-equal? #\b (+ bt 2))
          ;; (if (equal? (string-ref t (+ bt 1)) #\b)
          (if (traced-equal? #\b (+ bt 1))
              ;; (if (equal? (string-ref t (+ bt 0)) #\a)
              (if (traced-equal? #\a (+ bt 0))
                  bt
                  (loop-pc-1 bt lt 3))
              (let ([bt (+ bt 1)] [lt (- lt 1)])
                (cond
                 [(< lt 3)
                  -1]
                 ;; [(equal? (string-ref t (+ bt 0)) #\a)
                 [(traced-equal? #\a (+ bt 0))
                  (loop-pt-1 bt lt)]
                 [else
                  (loop-pc-1 bt lt 2)])))
          (let ([bt (+ bt 2)] [lt (- lt 2)])
            (cond
             [(< lt 3)
              -1]
             ;; [(equal? (string-ref t (+ bt 0)) #\a)
             [(traced-equal? #\a (+ bt 0))
              (loop-pt-1 bt lt)]
             [else
              (loop-pc-1 bt lt 1)]))))
    (let ([lt (string-length t)])
      (if (< lt 3)
          -1
          (loop-pt-1 0 lt))))
  (if (string=? pat "abb")
      (begin (match-abb txt)
             (reverse trace))
      '()))

(define (amtoft-al-Jones02-fig-3-abb-prune-duplicates pat txt)
  "Torben Amtoft, Charles Consel, Olivier Danvy, and Karoline
Malmkjær. The abstraction and instantiation of string-matching pro-
grams. In Torben Æ. Mogensen, David A. Schmidt, and I. Hal Sud-
borough, editors, The Essence of Computation: Complexity, Analysis,
Transformation. Essays Dedicated to Neil D. Jones, number 2566 in
Lecture Notes in Computer Science, pages 332–357. Springer, 2002."
  (define trace '())
  (define (traced-equal? char k) 
    (set! trace (cons k trace)) 
    (equal? char (string-ref txt k)))
  (define (match-abb t)
    (define (loop-pc-1 bt lt z)
      (let ([bt (+ bt z)] [lt (- lt z)])
        (if (< lt 3)
            -1
            (loop-pt-1 bt lt))))
    (define (loop-pt-1 bt lt)
      ;; (if (equal? (string-ref t (+ bt 2)) #\b)
      (if (traced-equal? #\b (+ bt 2))
          ;; (if (equal? (string-ref t (+ bt 1)) #\b)
          (if (traced-equal? #\b (+ bt 1))
              ;; (if (equal? (string-ref t (+ bt 0)) #\a)
              (if (traced-equal? #\a (+ bt 0))
                  bt
                  (loop-pc-1 bt lt 3))
              (let ([bt (+ bt 1)] [lt (- lt 1)])
                (cond
                 [(< lt 3)
                  -1]
                 [(equal? (string-ref t (+ bt 0)) #\a)
                 ;; [(traced-equal? #\a (+ bt 0))
                  (loop-pt-1 bt lt)]
                 [else
                  (loop-pc-1 bt lt 2)])))
          (let ([bt (+ bt 2)] [lt (- lt 2)])
            (cond
             [(< lt 3)
              -1]
             [(equal? (string-ref t (+ bt 0)) #\a)
             ;; [(traced-equal? #\a (+ bt 0))
              (loop-pt-1 bt lt)]
             [else
              (loop-pc-1 bt lt 1)]))))
    (let ([lt (string-length t)])
      (if (< lt 3)
          -1
          (loop-pt-1 0 lt))))
  (if (string=? pat "abb")
      (begin (match-abb txt)
             (reverse trace))
      '()))

(define (consel-danvy-IPL89-naive-approach pat txt)
  "Charles Consel and Olivier Danvy. Partial evaluation of pattern
matching in strings. Information Processing Letters, 30(2):79–86,
January 1989."
  (define (new-s s) (cons 0 s))
  (define index car)
  (define string cdr)
  (define trace '())
  (let ((null? (lambda (s)
                 (>= (index s) (string-length (string s)))))
        (car (lambda (s) s))
        (cdr (lambda (s) (cons (1+ (index s)) (string s))))
        (equal? (lambda (p t)
                  (set! trace (cons (index t) trace))
                  (equal? (string-ref (string p) (index p))
                          (string-ref (string t) (index t))))))
    (define (kmp p d)
      (or (null? p)
          (start p d)))
    (define (start p d)
      (and (not (null? d))
           (restart p d)))
    (define (restart p d)
      (if (equal? (car p) (car d))
          (loop (cdr p) (cdr d) p d)
          (start p (cdr d))))
    (define (loop p d pp dd)
      (or (null? p)
          (and (not (null? d))
               (if (equal? (car p) (car d))
                   (loop (cdr p) (cdr d) pp dd)
                   (restart pp (cdr dd))))))
    (kmp (new-s pat) (new-s txt))
    (reverse trace)))

(define (consel-danvy-IPL89-still-naive-approach pat txt)
  "Charles Consel and Olivier Danvy. Partial evaluation of pattern
matching in strings. Information Processing Letters, 30(2):79–86,
January 1989."
  (define (new-s s) (cons 0 s))
  (define index car)
  (define string cdr)
  (define trace '())
  (let ((null? (lambda (s)
                 (>= (index s) (string-length (string s)))))
        (car (lambda (s) s))
        (cdr (lambda (s) (cons (1+ (index s)) (string s))))
        (equal?pt (lambda (p t)
                    (set! trace (cons (index t) trace))
                    (equal? (string-ref (string p) (index p))
                            (string-ref (string t) (index t)))))
        (equal?pp (lambda (p pp)
                    (equal? (string-ref (string p) (index p))
                            (string-ref (string pp) (index pp)))))
        (length (lambda (s)
                  (- (string-length (string s)) (index s)))))
    (define (kmp p d)
      (or (null? p)
          (start p d)))
    (define (start p d)
      (and (not (null? d))
           (restart p d)))
    (define (restart p d)
      (if (equal?pt (car p) (car d))
          (or (null? (cdr p))
              (and (not (null? (cdr d)))
                   (loop (cdr p) (cdr d) p)))
          (start p (cdr d))))
    (define (loop p d pp)
      (if (equal?pt (car p) (car d))
          (or (null? (cdr p))
              (and (not (null? (cdr d)))
                   (loop (cdr p) (cdr d) pp)))
          (let ((np (static-kmp pp (cdr pp) (- (length (cdr pp)) (length p)))))
            (if (equal? np pp)
                (if (equal?pp (car pp) (car p))
                    (start pp (cdr d))
                    (restart pp d))
                (loop np d pp)))))
    (define (static-kmp p d n)
      (static-loop p d n p d n))
    (define (sub1 x) (- x 1))
    (define (static-loop p d n pp dd nn)
      (if (zero? n)
          p
          (if (equal?pp (car p) (car d))
              (static-loop (cdr p) (cdr d) (sub1 n) pp dd nn)
              (static-kmp pp (cdr dd) (sub1 nn)))))
    
    (kmp (new-s pat) (new-s txt))
    (reverse trace)))

(define (consel-danvy-IPL89-still-naive-ababc pat txt)
  "Charles Consel and Olivier Danvy. Partial evaluation of pattern
matching in strings. Information Processing Letters, 30(2):79–86,
January 1989."
  (define (new-s s) (cons 0 s))
  (define index car)
  (define string cdr)
  (define trace '())
  (let ((null? (lambda (s)
                 (>= (index s) (string-length (string s)))))
        (car (lambda (s) s))
        (cdr (lambda (s) (cons (1+ (index s)) (string s))))
        (equal? (lambda (char t)
                  (set! trace (cons (index t) trace))
                  (equal? char (string-ref (string t) (index t)))))
        (length (lambda (s)
                  (- (string-length (string s)) (index s)))))
    (define (kmp-0 d)
      (start-1 d))
    (define (start-1 d)
      (and (not (null? d))
           (restart-2 d)))
    (define (restart-2 d)
      (if (equal? #\a (car d))
          (and (not (null? d))
               (loop-3 (cdr d)))
          (start-1 (cdr d))))
    (define (loop-3 d)
      (if (equal? #\b (car d))
          (and (not (null? (cdr d)))
               (loop-4 (cdr d)))
          (restart-2 d)))
    (define (loop-4 d)
      (if (equal? #\a (car d))
          (and (not (null? (cdr d)))
               (loop-5 (cdr d)))
          (start-1 (cdr d))))
    (define (loop-5 d)
      (if (equal? #\b (car d))
          (and (not (null? (cdr d)))
               (loop-6 (cdr d)))
          (loop-3 d)))
    (define (loop-6 d)
      (or (equal? #\c (car d))
          (loop-4 d)))

    (if (string= "ababc" pat)
        (begin (kmp-0 (new-s txt))
               (reverse trace))
        '())))

(define (consel-danvy-IPL89-further-optimization pat txt)
  "Charles Consel and Olivier Danvy. Partial evaluation of pattern
matching in strings. Information Processing Letters, 30(2):79–86,
January 1989."
  (define (new-s s) (cons 0 s))
  (define index car)
  (define string cdr)
  (define trace '())
  (let ((null? (lambda (s)
                 (>= (index s) (string-length (string s)))))
        (car (lambda (s) s))
        (cdr (lambda (s) (cons (1+ (index s)) (string s))))
        (equal?pt (lambda (p t)
                    (set! trace (cons (index t) trace))
                    (equal? (string-ref (string p) (index p))
                            (string-ref (string t) (index t)))))
        (equal?pp (lambda (p pp)
                    (equal? (string-ref (string p) (index p))
                            (string-ref (string pp) (index pp)))))
        (length (lambda (s)
                  (- (string-length (string s)) (index s)))))
    (define (kmp p d)
      (or (null? p)
          (start p d)))
    (define (start p d)
      (and (not (null? d))
           (restart p d)))
    (define (restart p d)
      (if (equal?pt (car p) (car d))
          (or (null? (cdr p))
              (and (not (null? (cdr d)))
                   (loop (cdr p) (cdr d) p)))
          (start p (cdr d))))
    (define (loop p d pp)
      (if (equal?pt (car p) (car d))
          (or (null? (cdr p))
              (and (not (null? (cdr d)))
                   (loop (cdr p) (cdr d) pp)))
          (let ((np (static-kmp pp (cdr pp) (- (length (cdr pp)) (length p)))))
            (if (equal? np pp)
                (if (equal?pp (car pp) (car p))
                    (start pp (cdr d))
                    (restart pp d))
                (loop np d pp)))))
    (define (static-kmp p d n)
      (static-loop p d n p d n))
    (define (sub1 x) (- x 1))
    (define (static-loop p d n pp dd nn)
      (if (zero? n)
          (if (and (> nn 0)
                   (equal?pp (car p) (car d)))
              (static-kmp pp (cdr dd) (sub1 nn))
              p)
          (if (equal?pp (car p) (car d))
              (static-loop (cdr p) (cdr d) (sub1 n) pp dd nn)
              (static-kmp pp (cdr dd) (sub1 nn)))))
    
    (kmp (new-s pat) (new-s txt))
    (reverse trace)))

(define (consel-danvy-IPL89-further-optimization-abcabcacab pat txt)
  "Charles Consel and Olivier Danvy. Partial evaluation of pattern
matching in strings. Information Processing Letters, 30(2):79–86,
January 1989."
  (define (new-s s) (cons 0 s))
  (define index car)
  (define string cdr)
  (define trace '())
  (let ((null? (lambda (s)
                 (>= (index s) (string-length (string s)))))
        (car (lambda (s) s))
        (cdr (lambda (s) (cons (1+ (index s)) (string s))))
        (equal? (lambda (char t)
                  (set! trace (cons (index t) trace))
                  (equal? char (string-ref (string t) (index t)))))
        (length (lambda (s)
                  (- (string-length (string s)) (index s)))))
    (define (kmp-0 d) (start-1 d))
    (define (start-1 d)
      (and (not (null? d))
           (restart-2 d)))
    (define (restart-2 d)
      (if (equal? #\a (car d))
          (and (not (null? (cdr d)))
               (loop-3 (cdr d)))
          (start-1 (cdr d))))
    (define (loop-3 d)
      (if (equal? #\b (car d))
          (and (not (null? (cdr d)))
               (loop-4 (cdr d)))
          (restart-2 d)))
    (define (loop-4 d)
      (if (equal? #\c (car d))
          (and (not (null? (cdr d)))
               (loop-5 (cdr d)))
          (restart-2 d)))
    (define (loop-5 d)
      (if (equal? #\a (car d))
          (and (not (null? (cdr d)))
               (loop-6 (cdr d)))
          (start-1 (cdr d))))
    (define (loop-6 d)
      (if (equal? #\b (car d))
          (and (not (null? (cdr d)))
               (loop-7 (cdr d)))
          (restart-2 d)))
    (define (loop-7 d)
      (if (equal? #\c (car d))
          (and (not (null? (cdr d)))
               (loop-8 (cdr d)))
          (restart-2 d)))
    (define (loop-8 d)
      (if (equal? #\a (car d))
          (and (not (null? (cdr d)))
               (loop-9 (cdr d)))
          (start-1 (cdr d))))
    (define (loop-9 d)
      (if (equal? #\c (car d))
          (and (not (null? (cdr d)))
               (loop-10 (cdr d)))
          (loop-6 d)))
    (define (loop-10 d)
      (if (equal? #\a (car d))
          (and (not (null? (cdr d)))
               (loop-11 (cdr d)))
          (start-1 (cdr d))))
    (define (loop-11 d)
      (or (equal? #\b (car d))
          (restart-2 d)))

    (if (string= "abcabcacab" pat)
        (begin (kmp-0 (new-s txt))
               (reverse trace))
        '())))

(define (danvy-rohde-IPL06-sec-2 pat txt)
  "Olivier Danvy and Henning Korsholm Rohde. On obtaining the Boyer-
Moore string-matching algorithm by partial evaluation. Information
Processing Letters, 99(4):158–162, 2006."
  (define trace '())
  (define (equal-pt? j k) 
    (set! trace (cons k trace)) 
    (equal? (string-ref pat j) (string-ref txt k)))

  (define (sub1 x) (- x 1))
  (define (main p t)
    (match p t (sub1 (string-length p)) (sub1 (string-length p))))
  (define (match p t j k)
    (if (= j -1)
        (+ k 1)
        (if (>= k (string-length t))
            #f
            (compare p t j k))))
  (define (compare p t j k)
    (if (equal-pt? j k)
        (match p t (sub1 j) (sub1 k))
        (let ((offset (compute-offset p t j k)))
          (match p t (sub1 (string-length p)) (+ k offset)))))
  (define (compute-offset p t j k)
    (- (string-length p) j))
  (main pat txt)
  (reverse trace))

(define (danvy-rohde-IPL06-sec-3 pat txt)
  "Olivier Danvy and Henning Korsholm Rohde. On obtaining the Boyer-
Moore string-matching algorithm by partial evaluation. Information
Processing Letters, 99(4):158–162, 2006."
  (define trace '())
  (define (equal-pt? j k) 
    (set! trace (cons k trace)) 
    (equal? (string-ref pat j) (string-ref txt k)))
  (define (txt-ref k)
    (set! trace (cons k trace)) 
    (string-ref txt k))

  (define (sub1 x) (- x 1))
  (define (main p t)
    (match p t (sub1 (string-length p)) (sub1 (string-length p))))
  (define (match p t j k)
    (if (= j -1)
        (+ k 1)
        (if (>= k (string-length t))
            #f
            (compare p t j k))))
  (define (compare p t j k)
    (if (equal-pt? j k)
        (match p t (sub1 j) (sub1 k))
        (let ((offset (compute-offset p t j k)))
          (match p t (sub1 (string-length p)) (+ k offset)))))
  (define (compute-offset p t j k)
    (+ (- (string-length p) j 1) 
       (shift p (string-ref t (+ k (- (string-length p) j 1))))))
  (define (shift p c)
    (rematch p 1 c))
  (define (rematch p i c)
    (if (= i (string-length p))
        i
        (if (equal? c (string-ref p (- (string-length p) 1 i)))
            i
            (rematch p (1+ i) c))))
  (main pat txt)
  (reverse trace))

(define (danvy-rohde-IPL06-sec-3-aba pat txt)
  "Olivier Danvy and Henning Korsholm Rohde. On obtaining the Boyer-
Moore string-matching algorithm by partial evaluation. Information
Processing Letters, 99(4):158–162, 2006."
  (define trace '())
  (define (equal-with? char k)
    (set! trace (cons k trace)) 
    (equal? char (string-ref txt k)))
  (define (sub1 x) (- x 1))

  (define (main-aba t)
    (match-aba-2 t 2))
  (define (match-aba-2 t k)
    (if (>= k (string-length t))
        #f
        (if (equal-with? #\a k)
            (match-aba-1 t (sub1 k))
            (match-aba-2 t (+ k (shift-aba (string-ref t k)))))))
  (define (match-aba-1 t k)
    (if (>= k (string-length t))
        #f
        (if (equal-with? #\b k)
            (match-aba-0 t (sub1 k))
            (match-aba-2 t (+ k 1 (shift-aba (string-ref t (+ k 1))))))))
  (define (match-aba-0 t k)
    (if (>= k (string-length t))
        #f
        (if (equal-with? #\a k)
            (match-aba--1 t (sub1 k))
            (match-aba-2 t (+ k 2 (shift-aba (string-ref t (+ k 2))))))))
  (define (match-aba--1 t k)
    (+ k 1))
  (define (shift-aba x)
    (case x
      ((#\a) 2)
      ((#\b) 1)
      ((#\c) 3)
      (else (error "foo" x))))

  (if (string= pat "aba")
      (begin (main-aba txt)
             (reverse trace))
      '()))

(define (danvy-rohde-IPL06-sec-4 pat txt)
  "Olivier Danvy and Henning Korsholm Rohde. On obtaining the Boyer-
Moore string-matching algorithm by partial evaluation. Information
Processing Letters, 99(4):158–162, 2006."
  (define trace '())
  (define (equal-pt? j k) 
    (set! trace (cons k trace)) 
    (equal? (string-ref pat j) (string-ref txt k)))
  (define (txt-ref k)
    (set! trace (cons k trace)) 
    (string-ref txt k))

  (define pl (string-length pat))
  (define (sub1 x) (- x 1))
  (define (main p t)
    (match p t (sub1 pl) (sub1 pl)))
  (define (match p t j k)
    (if (= j -1)
        (+ k 1)
        (if (>= k (string-length t))
            #f
            (compare p t j k))))
  (define (compare p t j k)
    (if (equal-pt? j k)
        (match p t (sub1 j) (sub1 k))
        (let ((offset (compute-offset p t j k)))
          (match p t (sub1 pl) (+ k offset)))))
  (define (compute-offset p t j k)
    (+ (- pl j 1) 
       (max (rematch-gs p j (sub1 pl) (- pl 2))
            (+ (shift p (string-ref t k))
               (- pl) j 1))))
  (define (shift p c)
    (rematch p 1 c))
  (define (rematch p i c)
    (if (= i pl)
        i
        (if (equal? c (string-ref p (- pl 1 i)))
            i
            (rematch p (1+ i) c))))
  (define (rematch-gs p j j' k')
    (if (= k' -1)
        (+ j' 1)
        (if (= j j')
            (if (not (equal? (string-ref p j') (string-ref p k')))
                (- j' k')
                (rematch-gs p j (sub1 pl) (+ k' (- pl j' 2))))
            (if (equal? (string-ref p j') (string-ref p k'))
                (rematch-gs p j (sub1 j') (sub1 k'))
                (rematch-gs p j (sub1 pl) (+ k' (- pl j' 2)))))))

  (main pat txt)
  (reverse trace))

(define (knuth-morris-pratt-SIAM77-abcabcacab pat txt)
  "D. E. Knuth, J. H. Morris, and V. R. Pratt. Fast pattern matching in
strings. SIAM Journal on Computing, 6:323–350, 1977."
  (define trace '())
  (define (equal-with? char k)
    (set! trace (cons k trace)) 
    (equal? char (string-ref txt k)))

  (define (L0 k)
    (L1 (1+ k)))
  (define (L1 k)
    (if (not (equal-with? #\a k))
        (L0 k)
        (if (> (+ k 10) (string-length txt))
            #f
            (L2 (1+ k)))))
  (define (L2 k)
    (if (not (equal-with? #\b k))
        (L1 k)
        (L3 (1+ k))))
  (define (L3 k)
    (if (not (equal-with? #\c k))
        (L1 k)
        (L4 (1+ k))))
  (define (L4 k)
    (if (not (equal-with? #\a k))
        (L0 k)
        (L5 (1+ k))))
  (define (L5 k)
    (if (not (equal-with? #\b k))
        (L1 k)
        (L6 (1+ k))))
  (define (L6 k)
    (if (not (equal-with? #\c k))
        (L1 k)
        (L7 (1+ k))))
  (define (L7 k)
    (if (not (equal-with? #\a k))
        (L0 k)
        (L8 (1+ k))))
  (define (L8 k)
    (if (not (equal-with? #\c k))
        (L5 k)
        (L9 (1+ k))))
  (define (L9 k)
    (if (not (equal-with? #\a k))
        (L0 k)
        (L10 (1+ k))))
  (define (L10 k)
    (if (not (equal-with? #\b k))
        (L1 k)
        #t))
  (if (string=? pat "abcabcacab")
      (begin (L0 -1)
             (reverse trace))
      '()))
  
(define (queinnec-geffroy-WSA92-babar pat txt)
  "Christian Queinnec and Jean-Marie Geffroy. Partial evaluation applied
to symbolic pattern matching with intelligent backtrack. In Workshop
in Static Analysis, number 81–82 in Bigre, 1992."
  (define trace '())
  (define (equal-with? char k)
    (set! trace (cons k trace)) 
    (equal? char (string-ref txt k)))

  (define (new-s s) (cons 0 s))
  (define index car)
  (define string cdr)
  (let ((pair? (lambda (s) (<= (index s) (string-length (string s)))))
        (equal? (lambda (s char) 
                  (equal-with? char (index s))))
        (car (lambda (s) s))
        (cdr (lambda (s) (cons (+ (index s) 1) (string s))))
        (cddr (lambda (s) (cons (+ (index s) 2) (string s))))
        (cdddr (lambda (s) (cons (+ (index s) 3) (string s))))
        (caddddr (lambda (s) (cons (+ (index s) 4) (string s))))
        (cdddddr (lambda (s) (cons (+ (index s) 5) (string s)))))
    
    (define (match-0 E0)
      (match-4 E0))
    (define (match-4 z1)
      (if (pair? z1)
          (if (equal? (car z1) #\b)
              (match-12 (cdr z1) z1)
              (match-4 (cdr z1)))
          #f))
    (define (match-12 e0 z1)
      (if (pair? e0)
          (if (equal? (car e0) #\a)
              (let ((e3 (cdr e0)))
                (if (pair? e3)
                    (if (equal? (car e3) #\b)
                        (match-28 (cdr e3) z1)
                        (match-4 (cdddr z1)))
                    #f))
              (let ((e17 (cdr z1)))
                (match-146 (car e17) e17)))
          #f))
    (define (match-28 e0 z1)
      (if (pair? e0)
          (if (equal? (car e0) #\a)
              (let ((e3 (cdr e0)))
                (if (pair? e3)
                    (if (equal? (car e3) #\r)
                        #t
                        (if (equal? (caddddr z1) #\b)
                            (match-28 (cdddddr z1) (cddr z1))
                            (match-4 (cdddddr z1))))
                    #f))
              (let ((e35 (cdddr z1)))
                (match-146 (car e35) e35)))
          #f))
    (define (match-146 e0 z3)
      (if (equal? e0 #\b)
          (match-12 (cdr z3) z3)
          (match-4 (cdr z3))))

    (if (string=? pat "babar")
        (begin (match-0 (new-s txt))
               (reverse trace))
        '())))


(define (queinnec-geffroy-WSA92-foo pat txt)
  "Christian Queinnec and Jean-Marie Geffroy. Partial evaluation applied
to symbolic pattern matching with intelligent backtrack. In Workshop
in Static Analysis, number 81–82 in Bigre, 1992."
  (define trace '())
  (define (equal-with? char k)
    (set! trace (cons k trace)) 
    (equal? char (string-ref txt k)))

  (define (new-s s) (cons 0 s))
  (define index car)
  (define string cdr)
  (let ((pair? (lambda (s) (<= (index s) (string-length (string s)))))
        (equal? (lambda (s char) 
                  (equal-with? char (index s))))
        (car (lambda (s) s))
        (cdr (lambda (s) (cons (+ (index s) 1) (string s))))
        (cddr (lambda (s) (cons (+ (index s) 2) (string s))))
        (cdddr (lambda (s) (cons (+ (index s) 3) (string s))))
        (caddddr (lambda (s) (cons (+ (index s) 4) (string s))))
        (cdddddr (lambda (s) (cons (+ (index s) 5) (string s)))))

    (define (match-0 e0)
      (if (pair? e0)
          (let ((e1 (cdr e0)))
            (if (pair? e1)
                (let ((e2 (cdr e1)))
                  (if (pair? e2)
                      (let ((e3 (cdr e2)))
                        (if (equal? (car e2) #\o)
                            (if (equal? (car e1) #\o)
                                (if (equal? (car e0) #\f)
                                    #f
                                    (match-0 (cdddr e0)))
                                (let* ((e10 (cdr e0))
                                       (e11 (cdr e10)))
                                  (match-156 (cdr e11) e10)))
                            (let ((e14 (cddr e0)))
                              (match-216 (cdr e14) e14))))
                      #f))
                #f))
          #f))
    (define (match-216 e0 z4)
      (if (pair? e0)
          (let ((e5 (cdr e0)))
            (if (pair? e5)
                (let ((e6 (cdr e5)))
                  (if (equal? (car e5) #\o)
                      (if (equal? (car e0) #\o)
                          (if (equal? (car z4) #\f)
                              #t
                              (match-0 (cdddr z4)))
                          (let* ((e14 (cdr z4))
                                 (e15 (cdr e14)))
                            (match-156 (cdr e15) e14)))
                      (let ((e18 (cddr z4)))
                        (match-216 (cdr e18) e18))))
                #f))
          #f))
    (define (match-156 e0 z6)
      (if (pair? e0)
          (let ((e7 (cdr e0)))
            (if (equal? (car e0) #\o)
                (if (equal? (car z6) #\f)
                    #t
                    (let* ((e13 (cdr (cdr z6)))
                           (e14 (cdr e13)))
                      (match-0 e14)))
                (let ((e16 (cddr z6)))
                  (match-216 (cdr e16) e16))))
          #f))

    (if (string=? pat "foo")
        (begin (match-0 (new-s txt))
               (reverse trace))
        '())))

(define (soerensen-al-JFP96-fig-4-aab pat txt)
  "Morten Heine Sørensen, Robert Glueck, and Neil D. Jones.  A
positive supercompiler. Journal of Functional Programming,
6(6):811–838, 1996."
  (define trace '())
  (define (traced-equal? char s') 
    (set! trace (cons s' trace)) 
    (equal? char (string-ref txt s')))
  (define txt-length (string-length txt))
  (define (loop-aab ss)
    (if (>= ss txt-length)
        trace
        (let ((s' ss)
              (ss' (1+ ss)))
          (if (traced-equal? #\a s')
              (loop-ab ss')
              (loop-aab ss')))))
  (define (loop-ab ss)
    (if (>= ss txt-length)
        trace
        (let ((s' ss)
              (ss' (1+ ss)))
          (if (traced-equal? #\a s')
              (loop-b ss')
              (loop-aab ss')))))
  (define (loop-b ss)
    (if (>= ss txt-length)
        trace
        (let ((s' ss)
              (ss' (1+ ss)))
          (if (traced-equal? #\b s')
              trace
              (if (traced-equal? #\a s')
                  (loop-b ss')
                  (loop-aab ss'))))))
  (if (string= pat "aab")
      (begin (loop-aab 0)
             (reverse trace))
      '()))

(define (soerensen-al-JFP96-fig-11-aab pat txt)
  "Morten Heine Sørensen, Robert Glueck, and Neil D. Jones.  A
positive supercompiler. Journal of Functional Programming,
6(6):811–838, 1996."
  (define trace '())
  (define (traced-equal? char s') 
    (set! trace (cons s' trace)) 
    (equal? char (string-ref txt s')))
  (define txt-length (string-length txt))
  (define (loop-aab ss)
    (if (>= ss txt-length)
        #f
        (let ((s' ss)
              (ss' (1+ ss)))
          (if (traced-equal? #\a s')
              (loop-ab ss')
              (loop-aab ss')))))
  (define (loop-ab ss)
    (if (>= ss txt-length)
        #f
        (let ((s' ss)
              (ss' (1+ ss)))
          (if (traced-equal? #\a s')
              (loop-b ss')
              (if (traced-equal? #\a s')
                  (loop-ab ss')
                  (loop-aab ss'))))))
  (define (loop-b ss)
    (if (>= ss txt-length)
        #f
        (let ((s' ss)
              (ss' (1+ ss)))
          (if (traced-equal? #\b s')
              #t
              (if (traced-equal? #\a s')
                  (loop-b ss')
                  (if (traced-equal? #\a s')
                      (loop-ab ss')
                      (loop-aab ss')))))))
  (if (string= pat "aab")
      (begin (loop-aab 0)
             (reverse trace))
      '()))

(define (soerensen-al-JFP96-fig-18-fixed pat txt)
  "Morten Heine Sørensen, Robert Glueck, and Neil D. Jones.  A
positive supercompiler. Journal of Functional Programming,
6(6):811–838, 1996."
  (define trace '())
  (define (pat-pat-equal? pat-index pat-index')
    (equal? (string-ref pat pat-index) (string-ref pat pat-index')))
  (define (pat-txt-equal? pat-index txt-index) 
    (set! trace (cons txt-index trace)) 
    (equal? (string-ref pat pat-index) (string-ref txt txt-index)))
  (define (match pp ss)
    (loop pp ss pp '() '()))
  (define (loop pp ss op bb ob)
    (if (null? pp)
        #t
        (let ((p' (car pp))
              (pp' (cdr pp)))
          (if (null? bb)
              (if (null? ss)
                  #f
                  (let ((s' (car ss))
                        (ss' (cdr ss)))
                    (if (pat-txt-equal? p' s')
                        (loop pp' ss' op '() (append ob (list p')))
                        (next op ss ob))))
              (let ((b' (car bb))
                    (bb' (cdr bb)))
                (if (pat-pat-equal? p' b')
                    (loop pp' ss op bb' ob)
                    (loop op ss op (cdr ob) (cdr ob))))))))
  (define (next op ss ob)
    (if (null? ob)
        (loop op (cdr ss) op '() '())
        (let ((o' (car ob))
              (ob' (cdr ob)))
          (loop op ss op ob' ob'))))
  (let ((pp (order-left-to-right (string-length pat)))
        (ss (order-left-to-right (string-length txt))))
    (loop pp ss pp '() '())
    (reverse trace)))

(define (staged-zhu-takaoka pat txt)
  (define trace '())
  (define trace-matching-phase '())
  (define (equal-pt? j k) 
    (set! trace-matching-phase (cons k trace-matching-phase)) 
    (equal? (string-ref pat j) (string-ref txt k)))
  (define (txt-ref k)
    (set! trace-matching-phase (cons k trace-matching-phase)) 
    (string-ref txt k))
  (define (new-matching-phase)
    (set! trace (append trace
                        (trace-remove-duplicates 
                         (reverse trace-matching-phase))))
    (set! trace-matching-phase '()))

  (define pl (string-length pat))
  (define (sub1 x) (- x 1))
  (define (main p t)
    (match p t (sub1 pl) (sub1 pl)))
  (define (match p t j k)
    (if (= j -1)
        (+ k 1)
        (if (>= k (string-length t))
            #f
            (compare p t j k))))
  (define (compare p t j k)
    (if (equal-pt? j k)
        (match p t (sub1 j) (sub1 k))
        (let ((offset (compute-offset p t j k)))
          (new-matching-phase)
          (match p t (sub1 pl) (+ k offset)))))
  (define (compute-offset p t j k)
    (+ (- pl j 1)
       (max (rematch-gs p j (sub1 pl) (- pl 2))
            (let ((last-txt-pat-index (+ pl (- j) k (- 1))))
              (shift p (txt-ref (- last-txt-pat-index 1)) 
                     (txt-ref last-txt-pat-index))))))
  (define (shift p c1 c2)
    (if (equal? c2 (string-ref p 0))
        (min (- pl 1) (rematch p 1 c1 c2))
        (rematch p 1 c1 c2)))
  (define (rematch p i c1 c2)
    (if (= i pl)
        i
        (if (and (equal? c1 (string-ref p (- pl i 1)))
                 (equal? c2 (string-ref p (- pl i))))
            (- i 1)
            (rematch p (1+ i) c1 c2))))
  (define (rematch-gs p j j' k')
    (if (= k' -1)
        (+ j' 1)
        (if (= j j')
            (if (not (equal? (string-ref p j') (string-ref p k')))
                (- j' k')
                (rematch-gs p j (sub1 pl) (+ k' (- pl j' 2))))
            (if (equal? (string-ref p j') (string-ref p k'))
                (rematch-gs p j (sub1 j') (sub1 k'))
                (rematch-gs p j (sub1 pl) (+ k' (- pl j' 2)))))))

  (main pat txt)
  (new-matching-phase)
  trace)

"matchers.scm"