;; matchers.scm
;; 
;; This file defines lists of string matchers. These lists are applyed
;; to our methods of comparing string matchers.

(define ext-path (string-append (getcwd) "/bin/"))
(define dyn-path (string-append (getcwd) "/lib/")) ; needs absolute path

(define (ext command)
  (make-external-matcher 
   (string-append ext-path command)))

(define (dyn name)
  (make-dynamic-matcher
   (string-append dyn-path name)
   name))

(define framework-matchers
  `(
    ("fw_naive" ,naive)
    ("fw_mp" ,mp)
    ("fw_kmp" ,kmp)
    ("fw_horspool" ,horspool)
    ("fw_horspool-right-to-left" ,horspool-right-to-left)
    ("fw_boyer-moore" ,boyer-moore)
    ("fw_automaton" ,automaton)
    ("fw_optimal-bm" ,optimal-bm)
    ("fw_partsch-stomp" ,partsch-stomp)
    ("fw_original-bm" ,original-bm)
    ("fw_raita" ,raita)
    ("fw_quick-search" ,quick-search)
    ("fw_smith" ,smith)
    ("fw_not-so-naive" ,not-so-naive)
    ("fw_zhu-takaoka" ,zhu-takaoka)
    ))

(define (framework-permutations up-to)
  "Generate permutations of string matchers using our framework UP-TO
this number of positive and negative matching phases saved"
  (define (walk matchs' orders' poss' negs')
    (define (walk' matchs orders poss negs)
      (cond
       ((null? matchs)
        '())
       ((null? orders)
        (walk' (cdr matchs) orders' poss' negs'))
       ((null? poss)
        (walk' matchs (cdr orders) poss' negs'))
       ((null? negs)
        (walk' matchs orders (cdr poss) negs'))
       (else
        (cons (list (string-append "fw_" (caar matchs) 
                                   "_" (caar orders)
                                   "_" (caar poss)
                                   "_" (caar negs))
                    (make-matcher
                     ((cadar matchs) (cadar orders)
                      (compose (cadar poss) (cadar negs)))))
              (walk' matchs orders poss (cdr negs))))))
    (walk' matchs' orders' poss' negs'))
  (define (make-input-string s prune-x prune-x-older-than)
    `((,(string-append "0" s) ,prune-x)
      ,@(map (lambda (i) 
               (list (string-append (number->string i) s)
                     (prune-x-older-than i)))
             (range-up 1 (1+ up-to)))
      (,s ,prune-none)))
    
  (walk `(("no-tbl_skip" ,match-basic)
          ("no-tbl_no-skip" ,match-basic-shifts)
          ("tbl_skip" ,match-table)
          ("tbl_no-skip" ,match-table-shifts))
        `(("l2r" ,order-left-to-right)
          ("r2l" ,order-right-to-left))
        (make-input-string "pos" prune-pos prune-pos-older-than)
        (make-input-string "neg" prune-neg prune-neg-older-than)))


(define framework-permutations-2 (framework-permutations 2))

(define framework-permutations-2-interesting
  `(
    ("fw_no-tbl_skip_l2r_0pos_0neg" ,(make-matcher
                                      (match-basic order-left-to-right
                                                   (compose prune-pos
                                                            prune-neg))))
    ("fw_no-tbl_skip_l2r_pos_0neg" 
     ,(make-matcher
       (match-basic order-left-to-right
                    (compose prune-none
                             prune-neg))))
    ("fw_no-tbl_skip_l2r_pos_1neg" 
     ,(make-matcher
       (match-basic order-left-to-right
                    (compose prune-none
                             (prune-neg-older-than 1)))))
    ("fw_no-tbl_skip_l2r_pos_2neg" 
     ,(make-matcher
       (match-basic order-left-to-right
                    (compose prune-none
                             (prune-neg-older-than 2)))))
    ("fw_no-tbl_skip_l2r_pos_neg" 
     ,(make-matcher
       (match-basic order-left-to-right
                    (compose prune-none
                             prune-none))))
    ("fw_tbl_skip_l2r_pos_0neg" 
     ,(make-matcher
       (match-table order-left-to-right
                    (compose prune-none
                             prune-neg))))
    ("fw_no-tbl_skip_r2l_0pos_0neg" 
     ,(make-matcher
       (match-basic order-right-to-left
                    (compose prune-pos
                             prune-neg))))
    ("fw_tbl_skip_r2l_pos_0neg" 
     ,(make-matcher
       (match-table order-right-to-left
                    (compose prune-none
                             prune-neg))))
    ("fw_no-tbl_skip_r2l_pos_neg" 
     ,(make-matcher
       (match-basic order-right-to-left
                    (compose prune-none
                             prune-none))))
    ("fw_tbl_no-skip_r2l_1pos_0neg" 
     ,(make-matcher
       (match-table-shifts order-right-to-left
                           (compose (prune-pos-older-than 1)
                                    prune-neg))))
    ))

(define framework-permutations-2-distinct
  `(
    ("fw_no-tbl_skip_l2r_0pos_0neg" ,(make-matcher
                                      (match-basic order-left-to-right
                                                   (compose prune-pos
                                                            prune-neg))))
    ("fw_no-tbl_skip_l2r_pos_neg" ,(make-matcher
                                    (match-basic order-left-to-right
                                                 (compose prune-none
                                                          prune-none))))
    ("fw_no-tbl_skip_r2l_1pos_0neg" 
     ,(make-matcher
       (match-basic order-right-to-left
                    (compose (prune-pos-older-than 1)
                             prune-neg))))
    ("fw_no-tbl_skip_r2l_pos_1neg" 
     ,(make-matcher
       (match-basic order-right-to-left
                    (compose prune-none
                             (prune-neg-older-than 1)))))
    ("fw_no-tbl_skip_l2r_0pos_1neg" 
     ,(make-matcher
       (match-basic order-left-to-right
                    (compose prune-pos
                             (prune-neg-older-than 1)))))
    ("fw_no-tbl_no-skip_l2r_pos_0neg" 
     ,(make-matcher
       (match-basic-shifts order-left-to-right
                           (compose prune-none
                                    prune-neg))))
    ("fw_no-tbl_no-skip_l2r_pos_1neg" 
     ,(make-matcher
       (match-basic-shifts order-left-to-right
                           (compose prune-none
                                    (prune-neg-older-than 1)))))
    ("fw_no-tbl_skip_l2r_0pos_2neg" 
     ,(make-matcher
       (match-basic order-left-to-right
                    (compose prune-pos
                             (prune-neg-older-than 2)))))
    ("fw_no-tbl_skip_l2r_0pos_neg" 
     ,(make-matcher
       (match-basic order-left-to-right
                    (compose prune-pos
                             prune-none))))
    ("fw_no-tbl_no-skip_l2r_pos_neg" 
     ,(make-matcher
       (match-basic-shifts order-left-to-right
                           (compose prune-none
                                    prune-none))))
    ("fw_no-tbl_skip_l2r_pos_0neg" 
     ,(make-matcher
       (match-basic order-left-to-right
                    (compose prune-none
                             prune-neg))))
    ("fw_no-tbl_skip_l2r_pos_1neg" 
     ,(make-matcher
       (match-basic order-left-to-right
                    (compose prune-none
                             (prune-neg-older-than 1)))))
    ("fw_tbl_no-skip_l2r_pos_neg" 
     ,(make-matcher
       (match-table-shifts order-left-to-right
                           (compose prune-none
                                    prune-none))))
    ("fw_tbl_skip_l2r_pos_neg" 
     ,(make-matcher
       (match-table order-left-to-right
                    (compose prune-none
                             prune-none))))
    ("fw_no-tbl_skip_r2l_0pos_0neg" 
     ,(make-matcher
       (match-basic order-right-to-left
                    (compose prune-pos
                             prune-neg))))
    ("fw_no-tbl_no-skip_r2l_1pos_0neg" 
     ,(make-matcher
       (match-basic-shifts order-right-to-left
                           (compose (prune-pos-older-than 1)
                                    prune-neg))))
    ("fw_no-tbl_skip_r2l_2pos_0neg" 
     ,(make-matcher
       (match-basic order-right-to-left
                    (compose (prune-pos-older-than 2)
                             prune-neg))))
    ("fw_no-tbl_skip_r2l_pos_0neg" 
     ,(make-matcher
       (match-basic order-right-to-left
                    (compose prune-none
                             prune-neg))))
    ("fw_no-tbl_no-skip_r2l_2pos_0neg" 
     ,(make-matcher
       (match-basic-shifts order-right-to-left
                           (compose (prune-pos-older-than 2)
                                    prune-neg))))
    ("fw_no-tbl_no-skip_r2l_pos_0neg" 
     ,(make-matcher
       (match-basic-shifts order-right-to-left
                           (compose prune-none
                                    prune-neg))))
    ("fw_no-tbl_skip_r2l_0pos_1neg" 
     ,(make-matcher
       (match-basic order-right-to-left
                    (compose prune-pos
                             (prune-neg-older-than 1)))))
    ("fw_no-tbl_skip_r2l_1pos_1neg" 
     ,(make-matcher
       (match-basic order-right-to-left
                    (compose (prune-pos-older-than 1)
                             (prune-neg-older-than 1)))))
    ("fw_no-tbl_no-skip_r2l_1pos_1neg" 
     ,(make-matcher
       (match-basic-shifts order-right-to-left
                           (compose (prune-pos-older-than 1)
                                    (prune-neg-older-than 1)))))
    ("fw_no-tbl_no-skip_r2l_pos_1neg" 
     ,(make-matcher
       (match-basic-shifts order-right-to-left
                           (compose prune-none
                                    (prune-neg-older-than 1)))))
    ("fw_no-tbl_skip_r2l_pos_neg" 
     ,(make-matcher
       (match-basic order-right-to-left
                    (compose prune-none
                             prune-none))))
    ("fw_no-tbl_skip_r2l_1pos_neg" 
     ,(make-matcher
       (match-basic order-right-to-left
                    (compose (prune-pos-older-than 1)
                             prune-none))))
    ("fw_no-tbl_skip_r2l_0pos_2neg" 
     ,(make-matcher
       (match-basic order-right-to-left
                    (compose prune-pos
                             (prune-neg-older-than 2)))))
    ("fw_no-tbl_skip_r2l_0pos_neg" 
     ,(make-matcher
       (match-basic order-right-to-left
                    (compose prune-pos
                             prune-none))))
    ("fw_tbl_skip_r2l_pos_neg" 
     ,(make-matcher
       (match-table order-right-to-left
                    (compose prune-none
                             prune-none))))
    ("fw_tbl_skip_r2l_2pos_neg" 
     ,(make-matcher
       (match-table order-right-to-left
                    (compose (prune-pos-older-than 2)
                             prune-none))))
    ("fw_tbl_skip_r2l_1pos_neg" 
     ,(make-matcher
       (match-table order-right-to-left
                    (compose (prune-pos-older-than 1)
                             prune-none))))
    ("fw_tbl_no-skip_r2l_1pos_neg" 
     ,(make-matcher
       (match-table-shifts order-right-to-left
                           (compose (prune-pos-older-than 1)
                                    prune-none))))
    ("fw_tbl_no-skip_r2l_pos_neg" 
     ,(make-matcher
       (match-table-shifts order-right-to-left
                           (compose prune-none
                                    prune-none))))
    ("fw_no-tbl_no-skip_r2l_1pos_neg" 
     ,(make-matcher
       (match-basic-shifts order-right-to-left
                           (compose (prune-pos-older-than 1)
                                    prune-none))))
    ("fw_no-tbl_no-skip_r2l_pos_neg" 
     ,(make-matcher
       (match-basic-shifts order-right-to-left
                           (compose prune-none
                                    prune-none))))
    
    ))


(define trusted-handbook-matchers
  `(
    ("cl_naive" ,(dyn "brute"))
    ("cl_raita" ,(dyn "raita"))
    ("cl_horspool" ,(dyn "horspool"))
    ("cl_automaton" ,(dyn "automaton"))
    ("cl_knuth_morris_pratt" ,(dyn "knuth_morris_pratt"))
    ("cl_smith" ,(dyn "smith"))
    ("cl_boyer_moore" ,(dyn "boyer_moore"))
    ("cl_morris_pratt" ,(dyn "morris_pratt"))
    ("cl_not_so_naive" ,(dyn "not_so_naive"))
    ("cl_quick_search" ,(dyn "quick_search"))
    ("cl_zhu_takaoka" ,(dyn "zhu_takaoka"))
    ))

(define external-trusted-handbook-matchers
  '(
    ("cl_naive" ,(ext "brute"))
    ("cl_raita" ,(ext "raita"))
    ("cl_horspool" ,(ext "horspool"))
    ("cl_kmp_skip_search" ,(ext "kmp_skip_search"))
    ("cl_skip_search" ,(ext "skip_search"))
    ("cl_knuth_morris_pratt" ,(ext "knuth_morris_pratt"))
    ("cl_smith" ,(ext "smith"))
    ("cl_boyer_moore" ,(ext "boyer_moore"))
    ("cl_morris_pratt" ,(ext "morris_pratt"))
    ("cl_tuned_bm" ,(ext "tuned_bm"))
    ("cl_turbo_bm" ,(ext "turbo_bm"))
    ("cl_not_so_naive" ,(ext "not_so_naive"))
    ("cl_quick_search" ,(ext "quick_search"))
    ("cl_zhu_takaoka" ,(ext "zhu_takaoka"))
    ))

(define other-handbook-matchers
  `(
    ("cl_tuned_bm" ,(dyn "tuned_bm"))
    ("cl_turbo_bm" ,(dyn "turbo_bm"))
    ("cl_kmp_skip_search" ,(dyn "kmp_skip_search"))
    ("cl_skip_search" ,(dyn "skip_search"))
    ("cl_apostolico_crochemore" ,(dyn "apostolico_crochemore"))
    ("cl_galil_seiferas" ,(dyn "galil_seiferas"))
    ("cl_apostolico_giancarlo" ,(dyn "apostolico_giancarlo"))
    ("cl_reverse_colussi" ,(dyn "reverse_colussi"))
    ("cl_karp_rabin" ,(dyn "karp_rabin"))
    ("cl_reverse_factor" ,(dyn "reverse_factor"))
    ("cl_backward_oracle_matching" ,(dyn "backward_oracle_matching"))
    ("cl_berry_ravindran" ,(dyn "berry_ravindran"))
    ("cl_maximal_shift" ,(dyn "maximal_shift"))
    ("cl_colussi" ,(dyn "colussi"))
    ("cl_turbo_reverse_factor" ,(dyn "turbo_reverse_factor"))
    ("cl_forward_dawg_matching" ,(dyn "forward_dawg_matching"))
    ("cl_two_way" ,(dyn "two_way"))
    ("cl_galil_giancarlo" ,(dyn "galil_giancarlo"))
    ))

(define cited-matchers
  `(
    ("ager-al-2002-ASIA-PEPM02-fig-3" ,ager-al-2002-ASIA-PEPM02-fig-3)
    ("ager-al-2002-ASIA-PEPM02-fig-6" ,ager-al-2002-ASIA-PEPM02-fig-6)
    ("ager-al-TOPLAS06-fig-1" ,ager-al-TOPLAS06-fig-1)
    ("ager-al-TOPLAS06-fig-3" ,ager-al-TOPLAS06-fig-3)
    ("ager-al-TOPLAS06-fig-4" ,ager-al-TOPLAS06-fig-4)
    ("amtoft-al-Jones02-left-to-right" ,amtoft-al-Jones02-left-to-right)
    ("amtoft-al-Jones02-right-to-left" ,amtoft-al-Jones02-right-to-left)
    ("amtoft-al-Jones02-fig-2-aaa" ,amtoft-al-Jones02-fig-2-aaa)
    ("amtoft-al-Jones02-fig-3-abb" ,amtoft-al-Jones02-fig-3-abb)
    ("amtoft-al-Jones02-fig-3-abb-prune-duplicates" 
     ,amtoft-al-Jones02-fig-3-abb-prune-duplicates)
    ("consel-danvy-IPL89-naive-approach" ,consel-danvy-IPL89-naive-approach)
    ("consel-danvy-IPL89-still-naive-approach" 
     ,consel-danvy-IPL89-still-naive-approach)
    ("consel-danvy-IPL89-still-naive-ababc"
     ,consel-danvy-IPL89-still-naive-ababc)
    ("consel-danvy-IPL89-further-optimization"
     ,consel-danvy-IPL89-further-optimization)
    ("consel-danvy-IPL89-further-optimization-abcabcacab"
     ,consel-danvy-IPL89-further-optimization-abcabcacab)
    ("danvy-rohde-IPL06-sec-2" ,danvy-rohde-IPL06-sec-2)
    ("danvy-rohde-IPL06-sec-3" ,danvy-rohde-IPL06-sec-3)
    ("danvy-rohde-IPL06-sec-3-aba" ,danvy-rohde-IPL06-sec-3-aba)
    ("danvy-rohde-IPL06-sec-4" ,danvy-rohde-IPL06-sec-4)
    ("knuth-morris-pratt-SIAM77-abcabcacab"
     ,knuth-morris-pratt-SIAM77-abcabcacab)
    ("queinnec-geffroy-WSA92-babar"
     ,queinnec-geffroy-WSA92-babar)
    ("queinnec-geffroy-WSA92-foo"
     ,queinnec-geffroy-WSA92-foo)
    ("soerensen-al-JFP96-fig-4-aab" ,soerensen-al-JFP96-fig-4-aab)
    ("soerensen-al-JFP96-fig-11-aab" ,soerensen-al-JFP96-fig-11-aab)
    ("soerensen-al-JFP96-fig-18-fixed" ,soerensen-al-JFP96-fig-18-fixed)
    ))

(define matchers-chosen
  (append framework-matchers 
          trusted-handbook-matchers
          cited-matchers
          framework-permutations-2-interesting))

(define matchers-chosen-distinct-fully
  `( 
    ("fw_naive" ,naive)
    ("fw_mp" ,mp)
    ("fw_smith" ,smith)
    ("fw_automaton" ,automaton)
    ("fw_quick-search" ,quick-search)
    ("fw_not-so-naive" ,not-so-naive)
    ("fw_horspool" ,horspool)
    ("fw_raita" ,raita)
    ("fw_no-tbl_skip_r2l_0pos_0neg" 
     ,(make-matcher
       (match-basic order-right-to-left
                    (compose prune-pos
                             prune-neg))))
    ("fw_horspool-right-to-left" ,horspool-right-to-left)
    ("fw_no-tbl_skip_r2l_pos_neg" 
     ,(make-matcher
       (match-basic order-right-to-left
                    (compose prune-none
                             prune-none))))
    ("fw_optimal-bm" ,optimal-bm)
    ("fw_partsch-stomp" ,partsch-stomp)
    ("consel-danvy-IPL89-still-naive-approach" 
     ,consel-danvy-IPL89-still-naive-approach)
    ("fw_original-bm" ,original-bm)
    ("fw_boyer-moore" ,boyer-moore)
    ("fw_kmp" ,kmp)
    ("fw_no-tbl_skip_l2r_pos_neg" 
     ,(make-matcher
       (match-basic order-left-to-right
                    (compose prune-none
                             prune-none))))
    ("fw_zhu-takaoka" ,zhu-takaoka)
    ))

(define matchers-all
  (append framework-matchers 
          trusted-handbook-matchers
          other-handbook-matchers
          cited-matchers
          framework-permutations-2))

(define matchers-all-distinct
  `(
    ("cl_karp_rabin" ,(dyn "karp_rabin"))
    ("fw_naive" ,naive)
    ("cl_galil_seiferas" ,(dyn "galil_seiferas"))
    ("fw_no-tbl_skip_l2r_0pos_1neg" 
     ,(make-matcher
       (match-basic order-left-to-right
                    (compose prune-pos
                             (prune-neg-older-than 1)))))
    ("fw_no-tbl_no-skip_l2r_pos_0neg" 
     ,(make-matcher
       (match-basic-shifts order-left-to-right
                           (compose prune-none
                                    prune-neg))))
    ("fw_mp" ,mp)
    ("fw_smith" ,smith)
    ("fw_quick-search" ,quick-search)
    ("cl_berry_ravindran" ,(dyn "berry_ravindran"))
    ("fw_automaton" ,automaton)
    ("fw_not-so-naive" ,not-so-naive)
    ("cl_apostolico_crochemore" ,(dyn "apostolico_crochemore"))
    ("cl_maximal_shift" ,(dyn "maximal_shift"))
    ("cl_two_way" ,(dyn "two_way"))
    ("cl_turbo_reverse_factor" ,(dyn "turbo_reverse_factor"))
    ("cl_skip_search" ,(dyn "skip_search"))
    ("cl_kmp_skip_search" ,(dyn "kmp_skip_search"))
    ("cl_zhu_takaoka" ,(dyn "zhu_takaoka"))
    ("cl_turbo_bm" ,(dyn "turbo_bm"))
    ("fw_no-tbl_skip_r2l_1pos_0neg" 
     ,(make-matcher
       (match-basic order-right-to-left
                    (compose (prune-pos-older-than 1)
                             prune-neg))))
    ("fw_horspool" ,horspool)
    ("cl_tuned_bm" ,(dyn "tuned_bm"))
    ("fw_raita" ,raita)
    ("fw_tbl_skip_r2l_1pos_neg" 
     ,(make-matcher
       (match-table order-right-to-left
                    (compose (prune-pos-older-than 1)
                             prune-none))))
    ("cl_reverse_colussi" ,(dyn "reverse_colussi"))
    ("fw_no-tbl_no-skip_l2r_pos_1neg" 
     ,(make-matcher
       (match-basic-shifts order-left-to-right
                           (compose prune-none
                                    (prune-neg-older-than 1)))))
    ("fw_tbl_no-skip_l2r_pos_neg" 
     ,(make-matcher
       (match-table-shifts order-left-to-right
                           (compose prune-none
                                    prune-none))))
    ("fw_no-tbl_skip_l2r_pos_neg" 
     ,(make-matcher
       (match-basic order-left-to-right
                    (compose prune-none
                             prune-none))))
    ("cl_galil_giancarlo" ,(dyn "galil_giancarlo"))
    ("cl_colussi" ,(dyn "colussi"))
    ("cl_backward_oracle_matching" ,(dyn "backward_oracle_matching"))
    ("cl_reverse_factor" ,(dyn "reverse_factor"))
    ("fw_no-tbl_skip_r2l_0pos_0neg" 
     ,(make-matcher
       (match-basic order-right-to-left
                    (compose prune-pos
                             prune-neg))))
    ("fw_no-tbl_no-skip_r2l_1pos_0neg" 
     ,(make-matcher
       (match-basic-shifts order-right-to-left
                           (compose (prune-pos-older-than 1)
                                    prune-neg))))
    ("fw_no-tbl_skip_r2l_2pos_0neg" 
     ,(make-matcher
       (match-basic order-right-to-left
                    (compose (prune-pos-older-than 2)
                             prune-neg))))
    ("fw_no-tbl_skip_r2l_pos_0neg" 
     ,(make-matcher
       (match-basic order-right-to-left
                    (compose prune-none
                             prune-neg))))
    ("fw_no-tbl_no-skip_r2l_2pos_0neg" 
     ,(make-matcher
       (match-basic-shifts order-right-to-left
                           (compose (prune-pos-older-than 2)
                                    prune-neg))))
    ("fw_no-tbl_no-skip_r2l_pos_0neg" 
     ,(make-matcher
       (match-basic-shifts order-right-to-left
                           (compose prune-none
                                    prune-neg))))
    ("fw_no-tbl_skip_r2l_0pos_1neg" 
     ,(make-matcher
       (match-basic order-right-to-left
                    (compose prune-pos
                             (prune-neg-older-than 1)))))
    ("cl_apostolico_giancarlo" ,(dyn "apostolico_giancarlo"))
    ("fw_no-tbl_skip_r2l_1pos_1neg" 
     ,(make-matcher
       (match-basic order-right-to-left
                    (compose (prune-pos-older-than 1)
                             (prune-neg-older-than 1)))))
    ("fw_horspool-right-to-left" ,horspool-right-to-left)
    ("fw_partsch-stomp" ,partsch-stomp)
    ("fw_no-tbl_skip_r2l_1pos_neg" 
     ,(make-matcher
       (match-basic order-right-to-left
                    (compose (prune-pos-older-than 1)
                             prune-none))))
    ("fw_no-tbl_skip_r2l_pos_neg" 
     ,(make-matcher
       (match-basic order-right-to-left
                    (compose prune-none
                             prune-none))))
    ("fw_tbl_no-skip_r2l_pos_neg" 
     ,(make-matcher
       (match-table-shifts order-right-to-left
                           (compose prune-none
                                    prune-none))))
    ("fw_no-tbl_skip_l2r_0pos_2neg" 
     ,(make-matcher
       (match-basic order-left-to-right
                    (compose prune-pos
                             (prune-neg-older-than 2)))))
    ("fw_no-tbl_skip_l2r_0pos_neg" 
     ,(make-matcher
       (match-basic order-left-to-right
                    (compose prune-pos
                             prune-none))))
    ("fw_no-tbl_no-skip_l2r_pos_neg" 
     ,(make-matcher
       (match-basic-shifts order-left-to-right
                           (compose prune-none
                                    prune-none))))
    ("consel-danvy-IPL89-still-naive-approach" 
     ,consel-danvy-IPL89-still-naive-approach)
    ("fw_kmp" ,kmp)
    ("fw_no-tbl_no-skip_r2l_1pos_1neg" 
     ,(make-matcher
       (match-basic-shifts order-right-to-left
                           (compose (prune-pos-older-than 1)
                                    (prune-neg-older-than 1)))))
    ("fw_no-tbl_no-skip_r2l_pos_1neg" 
     ,(make-matcher
       (match-basic-shifts order-right-to-left
                           (compose prune-none
                                    (prune-neg-older-than 1)))))
    ("fw_original-bm" ,original-bm)
    ("fw_boyer-moore" ,boyer-moore)
    ("fw_optimal-bm" ,optimal-bm)
    ("fw_tbl_skip_r2l_2pos_neg" 
     ,(make-matcher
       (match-table order-right-to-left
                    (compose (prune-pos-older-than 2)
                             prune-none))))
    ("fw_no-tbl_skip_r2l_0pos_2neg" 
     ,(make-matcher
       (match-basic order-right-to-left
                    (compose prune-pos
                             (prune-neg-older-than 2)))))
    ("fw_no-tbl_skip_r2l_0pos_neg" 
     ,(make-matcher
       (match-basic order-right-to-left
                    (compose prune-pos
                             prune-none))))
    ("fw_no-tbl_no-skip_r2l_1pos_neg" 
     ,(make-matcher
       (match-basic-shifts order-right-to-left
                           (compose (prune-pos-older-than 1)
                                    prune-none))))
    ("fw_no-tbl_no-skip_r2l_pos_neg" 
     ,(make-matcher
       (match-basic-shifts order-right-to-left
                           (compose prune-none
                                    prune-none))))
    ))
   
(define matchers-perms framework-permutations-2)

"matchers.scm"
