;;;; matchers.scm example file
;; Defines set of matchers to use when applying the framework
;; comparison methods. Definitions referenced here are defined in
;; framework/matchers.scm
;;
;; the "fw" in "fw_naive" stands for framework
;;
;; the "cl" in "cl_naive" stands for Charras and Lecroq 
;; (the authors of the Handbook of Exact String-Matching Algorithms)
;;
;; ext (external) matchers are programs called from the
;; shell with 2 arguments (the pattern and the shell). These programs
;; return a space seperated list of integers signifying the trace.
;; test-case for (ext "brute")
;;   $ ./bin/brute abaa abbabacabaa
;;   0 1 2 1 2 3 4 5 6 4 5 6 6 7 8 9 10
;;
;; dyn (dynamic) matchers are functions loaded from a dynamic
;; library. These libraries contain a function called 'init' that
;; initializes a function with the same of the second argument (if
;; dyn1) or the first argument if dyn.
;;
;; (dyn "morris_pratt") tries to load:
;;
;;    ./lib/morris_pratt.so
;;
;; There is assumed to be a function called morris_pratt in
;; morris_pratt.so. This is of the type:
;;
;;    List morris_pratt(char* pattern, int pattern_size,
;;                      char* text, int text_size)
;;
;; Where List is a linked list defined in c_matchers/linkedlist.h as:
;;
;;    struct _cell{
;;      int element; 
;;      struct _cell *next;
;;    };
;;    typedef struct _cell *List;
;;
;; The linked list holds the trace of the matcher.


;; Example of user-defined list of string matchers
(define matchers-kmp-like
  `(
    ("Naive" ,naive)
    ("Morris-Pratt" ,mp)
    ("Knuth-Morris-Pratt" ,kmp)
    ("KMP-2neg" ,(make-matcher
                  (match-basic order-left-to-right 
                               (prune-neg-older-than 2))))
    ("Optimal-KMP" ,(make-matcher
                      (match-basic order-left-to-right prune-none)))
    ))

;; Defined lists of string matchers
matchers-chosen ; Chosen matchers (from Chapter 4 - Results)
matchers-chosen-distinct-fully ; Matchers used for generating
                               ; evolutionay trees (none are
                               ; trace-equivalent, none are
                               ; partially-defined (specialized))
matchers-all
matchers-all-distinct
cited-matchers
matchers-kmp-like

;; Last list is the one used
matchers-chosen
